from __future__ import annotations

import contextlib
import enum
import sys
from asyncio import CancelledError
from collections.abc import Callable
from collections.abc import Coroutine
from typing import Any
from typing import BinaryIO

import lldb

import pwndbg
from pwndbg.dbg_mod.lldb import YieldContinue
from pwndbg.dbg_mod.lldb import YieldSingleStep
from pwndbg.dbg_mod.lldb.repl import UserCancelledError
from pwndbg.dbg_mod.lldb.repl import print_info
from pwndbg.dbg_mod.lldb.repl.io import IODriver
from pwndbg.dbg_mod.lldb.repl.io import IODriverPlainText


class EventHandler:
    """
    The event types that make sense for us to track in the process driver aren't
    the same as the ones in the rest of Pwndbg, so we just expose the native
    events in process driver, and let the rest of the REPL deal with any
    complexities that might arise from the translation.

    This is mostly intended to keep the complexity of generating the START and
    NEW_THREAD events correctly out of the process driver.
    """

    def created(self):
        """
        This function is called when a process is created or attached to.
        """
        pass

    def suspended(self, cause: lldb.SBEvent):
        """
        This function is called when the execution of a process is suspended.
        """
        pass

    def resumed(self):
        """
        This function is called when the execution of a process is resumed.
        """
        pass

    def exited(self):
        """
        This function is called when a process terminates or is detached from.
        """
        pass

    def modules_loaded(self):
        """
        This function is called when a new modules have been loaded.
        """
        pass


class _PollResult:
    """
    Base class for results of the run loop.
    """

    pass


class _PollResultTimedOut(_PollResult):
    """
    Indicates the run loop has timed out.
    """

    __match_args__ = ("last_event",)

    def __init__(self, last_event: lldb.SBEvent | None):
        self.last_event = last_event


class _PollResultStopped(_PollResult):
    """
    Indicates that the process has stopped normally.
    """

    __match_args__ = ("event",)

    def __init__(self, event: lldb.SBEvent):
        self.event = event


class _PollResultExited(_PollResult):
    """
    Run loop result for when a process has terminated.
    """

    __match_args__ = ("status",)

    def __init__(self, status: int):
        self.status = status


class LaunchResult:
    """
    Base class for results of launch operations.
    """

    pass


class LaunchResultSuccess(LaunchResult):
    """
    Indicates that the process was fully launched or attached to.
    """

    pass


class LaunchResultEarlyExit(LaunchResult):
    """
    Indicates that the process was fully launched or attached to, but that it
    exited immediately, with no stop events.
    """

    pass


class LaunchResultConnected(LaunchResult):
    """
    Indicates that there has been a successful connection to a remote
    debugserver, but that no process is being debugged yet.
    """

    pass


class LaunchResultError(LaunchResult):
    """
    Indicates that there was an error launching the process.
    """

    __match_args__ = ("what", "disconnected")

    def __init__(self, what: lldb.SBError, disconnected: bool):
        self.what = what
        self.disconnected = disconnected


def _updates_scope_counter(target: str) -> Callable[[Callable[..., Any]], Any]:
    """
    ProcessDriver makes use of scope counters as part of the decision process
    for how and when cancellations should be raised. This decorator
    automatically updates a given scope counter.
    """

    def sub0(fn: Callable[..., Any]) -> Callable[..., Any]:
        def sub1(self: ProcessDriver, *args, **kwargs):
            setattr(self, target, getattr(self, target) + 1)
            try:
                self.debug_print(f"self.{target} += 1 ({getattr(self, target)})")
                return fn(self, *args, **kwargs)
            finally:
                setattr(self, target, getattr(self, target) - 1)
                self.debug_print(f"self.{target} -= 1 ({getattr(self, target)})")

        return sub1

    return sub0


class _IODriverState(enum.Enum):
    STOPPED = 0
    "The IODriver is not running, and must be entirely restarted"
    RUNNING = 1
    "The IODriver is running"
    PAUSED = 2
    "The IODriver is not running, but is alive and can be resumed"


class ProcessDriver:
    """
    Drives the execution of a process, responding to its events and handling its
    I/O, and exposes a simple synchronous interface to the REPL interface.

    # Signal Handler Safety
    With few exceptions, no method in this class is safe to call during signal
    handlers. Exceptions are always explicitly noted as such.

    # IODriver State Machine
    Because LLDB can make Python code from Pwndbg execute while an I/O driver is
    active, and having the I/O driver active while Pwndbg is running leads to
    all sorts of fun failure modes, we want to be able to pause it temporarily.

    We, thus, use the states described in _IODriverState to keep track of what
    operations may be performed on the current IODriver.
    """

    io: IODriver
    process: lldb.SBProcess
    listener: lldb.SBListener
    debug: bool
    eh: EventHandler

    _hold_cancellation: int
    "Nested scope counter for cancellation suspension"

    _pending_cancellation: bool
    "Whether we should fire a cancellation request once we resume cancellations"

    _in_run_until_next_stop: int
    "Nested scope counter for _run_until_next_stop"

    _in_run_coroutine: int
    "Nested scope counter for run_coroutine"

    _io_driver_state: _IODriverState
    "Whether the I/O driver is currently running"

    _is_core_file: bool
    "Whether the current live process is actually a core file"

    def __init__(self, event_handler: EventHandler, debug=False):
        self.io = None
        self.process = None
        self.listener = None
        self.debug = debug
        self.eh = event_handler
        self._hold_cancellation = 0
        self._pending_cancellation = False
        self._in_run_until_next_stop = 0
        self._in_run_coroutine = 0
        self._io_driver_state = _IODriverState.STOPPED
        self._is_core_file = False

    def __enter__(self) -> ProcessDriver:
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb) -> None:
        if self.io is not None:
            self.io.close()

    def debug_print(self, *args, **kwargs) -> None:
        if self.debug:
            from io import StringIO

            with StringIO() as out:
                print("[*] ProcessDriver: ", end="", file=out)
                print(*args, file=out, **kwargs)

                message = out.getvalue().encode(sys.stdout.encoding)

            while len(message) > 0:
                try:
                    message = message[sys.stdout.buffer.write(message) :]
                    sys.stdout.buffer.flush()
                except BlockingIOError as e:
                    message = message[e.characters_written :]

    def has_process(self) -> bool:
        """
        Whether there's an active process in this driver.
        """
        return self.process is not None and self.process.GetState() != lldb.eStateConnected

    def has_connection(self) -> bool:
        """
        Whether this driver's connected to a target. All drivers that have an
        active process also must necessarily be connected.
        """
        return self.process is not None

    def is_core_file(self) -> bool:
        """
        Whether this driver's process is a core file.

        Core file processes don't support changes to the lifetime of the process.
        """
        return self._is_core_file

    def kill(self) -> lldb.SBError:
        """
        Kills the currently running process.
        """
        assert self.has_process(), "called kill() on a driver with no process"
        with self._suspend_interrupts():
            e = self.process.Kill()
            if not e.success:
                return e

            # Callers expect the process to have already been dropped by the
            # time this method returns, so we must drive the process to
            # completion.
            #
            # By the time we start driving, there might be pending stop events
            # that aren't necessarily the ones singnaling the termination of
            # the inferior. We keep calling _run_until_next_stop until it sees
            # that the process has been killed, which also clears driver state
            # for us automatically.
            while self.has_process():
                self._run_until_next_stop()

            return lldb.SBError()

    def interrupt(self, in_lldb_command_handler: bool = False) -> None:
        """
        Interrupts the currently running process or command.

        This method may be called during a signal handler.
        """
        if not self.has_process():
            return

        if self._in_run_until_next_stop > 0:
            if self._in_run_coroutine > 0:
                self.debug_print("Sending interrupt and requesting current coroutine stop")
            else:
                self.debug_print("Sending interrupt")

            # If we're in a coroutine, we should tell it to stop as soon as it gets out of _run_until_next_stop()
            self._pending_cancellation = self._in_run_coroutine > 0

            self.process.SendAsyncInterrupt()
        elif self._hold_cancellation > 0 and not in_lldb_command_handler:
            self.debug_print("Pushing pending cancellation")
            self._pending_cancellation = True
        else:
            self.debug_print(
                "Requesting cancellation immediately",
                "(forced by being in a command handler)" if self._hold_cancellation > 0 else "",
            )

            # This happens even if interrupts are suspended, if we're inside a
            # command handler. We shouldn't interrupt LLDB until it starts
            # executing our command handler, but we still want to be able to
            # interrupt the handler if it's a particularly long-running command
            # like `search`.
            self._pending_cancellation = False
            raise UserCancelledError("user-requested cancellation")

    @contextlib.contextmanager
    def _suspend_interrupts(self, interrupt: Callable[[], None] | None = None):
        """
        Sometimes it's necessary to guard against interruption by
        self.interrupt, especially when being interrupted would lead to bad
        process state.
        """

        self._hold_cancellation += 1
        try:
            self.debug_print("Temporarily suspending cancellations")

            yield None
        finally:
            self.debug_print("Resuming cancellations")

            self._hold_cancellation -= 1
            if self._hold_cancellation == 0 and self._pending_cancellation:
                self.debug_print(
                    "Executing pending cancellation",
                )
                self._pending_cancellation = False
                if interrupt is None:
                    # The default action is to raise an exception in place.
                    raise UserCancelledError("user-requested cancellation")

                # Perform the called-provided action.
                interrupt()

    def _start_io_driver(self):
        """
        Starts the IODriver handling I/O from the process.
        """
        assert self._io_driver_state == _IODriverState.STOPPED
        self.io.start(process=self.process)

        self.debug_print(f"{self._io_driver_state} -> {_IODriverState.RUNNING}")
        self._io_driver_state = _IODriverState.RUNNING

    def _stop_io_driver(self):
        """
        Stops the IODriver handling I/O from the process.
        """
        assert (
            self._io_driver_state == _IODriverState.PAUSED
            or self._io_driver_state == _IODriverState.RUNNING
        )

        if self._io_driver_state == _IODriverState.PAUSED:
            # Not invalid, but currently a NOP. See pause_io_if_running()
            self._io_driver_state = _IODriverState.STOPPED
            return

        self.io.stop()

        self.debug_print(f"{self._io_driver_state} -> {_IODriverState.STOPPED}")
        self._io_driver_state = _IODriverState.STOPPED

    def pause_io_if_running(self) -> None:
        """
        Pauses the handling of process I/O if it is currently running.
        """
        if self._io_driver_state != _IODriverState.RUNNING:
            return

        # Currently, pausing and stopping both stop the IODriver. Nonetheless,
        # these operations must stay separate in the state machine, as they are
        # meaningfully distinct in other ways.
        self.io.stop()

        self.debug_print(f"{self._io_driver_state} -> {_IODriverState.PAUSED}")
        self._io_driver_state = _IODriverState.PAUSED

    def resume_io_if_running(self) -> None:
        """
        Resumes the handling of process I/O if it is currently running.
        """
        if self._io_driver_state != _IODriverState.PAUSED:
            return

        assert self.process is not None
        self.io.start(process=self.process)

        self.debug_print(f"{self._io_driver_state} -> {_IODriverState.RUNNING}")
        self._io_driver_state = _IODriverState.RUNNING

    @_updates_scope_counter(target="_in_run_until_next_stop")
    def _run_until_next_stop(
        self,
        with_io: bool = True,
        timeout: int = 1,
        first_timeout: int = 1,
        only_if_started: bool = False,
        fire_events: bool = True,
    ) -> _PollResult:
        """
        Runs the event loop of the process until the next stop event is hit, with
        a configurable timeouts for the first and subsequent timeouts.

        Optionally runs the I/O system alongside the event loop.

        If `only_if_started` is passed, this method will stop after the first
        timeout if it can't observe a state change to a running state, and I/O
        will only start running after the start event is observed.
        """

        assert not self._is_core_file, "_run_until_next_stop is not valid for core files"

        # If `only_if_started` is set, we defer the starting of the I/O driver
        # to the moment the start event is observed. Otherwise, we just start it
        # immediately.
        try:
            if with_io and not only_if_started:
                self._start_io_driver()

            # Pick the first timeout value.
            timeout_time = first_timeout

            # If `only_if_started` is not set, assume the process must have been
            # started by a previous action and is running.
            running = not only_if_started

            result = None
            last_event = None
            delay_until_io_stopped: list[Callable[[], None]] = []
            while True:
                event = lldb.SBEvent()
                if not self.listener.WaitForEvent(timeout_time, event):
                    self.debug_print(
                        f"Timed out after {timeout_time}s",
                    )
                    timeout_time = timeout

                    # If the process isn't running, we should stop.
                    if not running:
                        self.debug_print(
                            "Waited too long for process to start running, giving up",
                        )
                        result = _PollResultTimedOut(last_event)
                        break

                    continue
                last_event = event

                if self.debug:
                    descr = lldb.SBStream()
                    if event.GetDescription(descr):
                        self.debug_print(descr.GetData())
                    else:
                        self.debug_print(f"No description for {event}")

                if lldb.SBTarget.EventIsTargetEvent(event):
                    if event.GetType() == lldb.SBTarget.eBroadcastBitModulesLoaded:
                        # Notify the event handler that new modules got loaded in.
                        if fire_events:
                            self.eh.modules_loaded()

                elif lldb.SBProcess.EventIsProcessEvent(event):
                    if (
                        event.GetType() == lldb.SBProcess.eBroadcastBitSTDOUT
                        or event.GetType() == lldb.SBProcess.eBroadcastBitSTDERR
                    ):
                        # Notify the I/O driver that the process might have something
                        # new for it to consume.
                        self.io.on_output_event()
                    elif event.GetType() == lldb.SBProcess.eBroadcastBitStateChanged:
                        # The state of the process has changed.
                        new_state = lldb.SBProcess.GetStateFromEvent(event)
                        was_resumed = lldb.SBProcess.GetRestartedFromEvent(event)

                        if new_state == lldb.eStateStopped and not was_resumed:
                            # The process has stopped, so we're done processing events
                            # for the time being. Trigger the stopped event and return.
                            if fire_events:
                                self.eh.suspended(event)
                            result = _PollResultStopped(event)
                            break

                        if new_state == lldb.eStateRunning or new_state == lldb.eStateStepping:
                            running = True
                            # Start the I/O driver here if its start got deferred
                            # because of `only_if_started` being set.
                            if only_if_started and with_io:
                                self._start_io_driver()

                        if (
                            new_state == lldb.eStateExited
                            or new_state == lldb.eStateCrashed
                            or new_state == lldb.eStateDetached
                        ):
                            # Nothing else for us to do here. Clear our internal
                            # references to the process, fire the exit event, and leave.
                            self.debug_print(
                                f"Process exited with state {new_state}",
                            )
                            self.process = None
                            self.listener = None

                            if fire_events:
                                self.eh.exited()

                            if new_state == lldb.eStateExited:
                                proc = lldb.SBProcess.GetProcessFromEvent(event)
                                desc = (
                                    ""
                                    if not proc.exit_description
                                    else f" ({proc.exit_description})"
                                )
                                delay_until_io_stopped.append(
                                    lambda: print_info(
                                        f"process exited with status {proc.exit_state}{desc}"
                                    )
                                )
                            elif new_state == lldb.eStateCrashed:
                                delay_until_io_stopped.append(lambda: print_info("process crashed"))
                            elif new_state == lldb.eStateDetached:
                                delay_until_io_stopped.append(
                                    lambda: print_info("process detached")
                                )

                            result = _PollResultExited(new_state)
                            break
        finally:
            if self._io_driver_state != _IODriverState.STOPPED:
                self._stop_io_driver()
            if isinstance(result, _PollResultExited):
                self.io.close()
                self.io = None

        for fn in delay_until_io_stopped:
            fn()

        return result

    def cont(self) -> None:
        """
        Continues execution of the process this object is driving, and returns
        whenever the process stops.
        """
        assert not self._is_core_file, "cont is not valid for core files"
        assert self.has_process(), "called cont() on a driver with no process"

        self.eh.resumed()
        self.process.Continue()
        self._run_until_next_stop()

    def run_lldb_command(self, command: str, target: BinaryIO) -> None:
        """
        Runs the given LLDB command and ataches I/O if necessary.
        """
        assert self.has_process(), "called run_lldb_command() on a driver with no process"

        with self._suspend_interrupts():
            ret = lldb.SBCommandReturnObject()
            self.process.GetTarget().GetDebugger().GetCommandInterpreter().HandleCommand(
                command, ret
            )

            if ret.IsValid():
                # LLDB can give us strings that may fail to encode.
                out = ret.GetOutput().strip()
                if len(out) > 0:
                    target.write(out.encode(sys.stdout.encoding, errors="backslashreplace"))
                    target.write(b"\n")
                out = ret.GetError().strip()
                if len(out) > 0:
                    target.write(out.encode(sys.stdout.encoding, errors="backslashreplace"))
                    target.write(b"\n")

                self.debug_print(f"LLDB Command finished with status: {ret.GetStatus():#x}")

                # Only call _run_until_next_stop() if the command started the process.
                start_expected = True
                s = ret.GetStatus()
                if s == lldb.eReturnStatusFailed:
                    start_expected = False
                if s == lldb.eReturnStatusQuit:
                    start_expected = False
                if s == lldb.eReturnStatusSuccessFinishResult:
                    start_expected = False
                if s == lldb.eReturnStatusSuccessFinishNoResult:
                    start_expected = False

                # It's important to note that we can't trigger the resumed event
                # now because the process might've already started, and LLDB
                # will fail to do most of the operations that we need while the
                # process is running. Ideally, we'd have a way to trigger the
                # event right before the process is resumed, but as far as I know,
                # there is no way to do that.
                #
                # TODO/FIXME: Find a way to trigger the continued event before the process is resumed in LLDB

                if self._is_core_file:
                    # LLDB simply doesn't support continuing core file processes.
                    # If we get here, either LLDB has completely lost it, or we
                    # don't know how to handle this kind of command. Bail.
                    assert not start_expected, "LLDB seemingly continued a core file process"

                    # Because we don't have an event source, we can't directly
                    # know if the process has been killed by the user. So, we
                    # manually check against the state of the process after
                    # every command.
                    self.debug_print(
                        f"Core File: Process state after command is {self.process.state:#x}"
                    )
                    if not self.process.IsValid() or self.process.state == lldb.eStateExited:
                        self.process = None
                        self._is_core_file = False
                        self.debug_print("exited()")
                        self.eh.exited()
                        self.debug_print("Core File: Finished debugging core file")

                elif not start_expected:
                    # Even if the current process has not been resumed, LLDB might
                    # have posted process events for us to handle. Do so now, but
                    # stop right away if there is nothing for us in the listener.
                    #
                    # This lets us handle things like `detach`, which would otherwise
                    # pass as a command that does not change process state at all.
                    self._run_until_next_stop(first_timeout=0, only_if_started=True)
                else:
                    self._run_until_next_stop()

    def run_coroutine(self, coroutine: Coroutine[Any, Any, None]) -> bool:
        """
        Runs the given coroutine and allows it to control the execution of the
        process in this driver. Returns `True` if the coroutine ran to completion,
        and `False` if it was cancelled.
        """
        assert not self._is_core_file, "run_coroutine is not valid for core files"
        try:
            return self._run_coroutine(coroutine)
        except CancelledError:
            # We got cancelled somewhere else.
            return False

    @_updates_scope_counter("_in_run_coroutine")
    def _run_coroutine(self, coroutine: Coroutine[Any, Any, None]) -> bool:
        """
        This loop may be spuriously cancelled. We handle that in run_coroutine().
        """

        assert not self._is_core_file, "_run_coroutine is not valid for core files"
        assert self.has_process(), "called run_coroutine() on a driver with no process"

        self.debug_print("Coroutine: Starting")
        exceptions: list[BaseException] = []

        def queue_cancel():
            self.debug_print("Coroutine: Queueing up user cancellation exception")
            exceptions.append(UserCancelledError())

        while True:
            try:
                if len(exceptions) == 0:
                    self.debug_print("Coroutine: Resuming")
                    step = coroutine.send(None)
                else:
                    self.debug_print(f"Coroutine: Raising {type(exceptions[-1])}")
                    step = coroutine.throw(exceptions[-1])
                    # The coroutine has caught the exception. Continue running
                    # it as if nothing happened.
                    exceptions.pop()
            except StopIteration:
                # We got to the end of the coroutine. We're done.
                self.debug_print("Coroutine: Ran to completion")
                break
            except CancelledError:
                # We requested that the coroutine be cancelled, and it didn't
                # override our decision. We're done.
                self.debug_print("Coroutine: Cancelled")
                break

            # Being interrupted here would be bad for keeping the state of the
            # process consistent.
            with self._suspend_interrupts(interrupt=queue_cancel):
                if isinstance(step, YieldSingleStep):
                    self.debug_print("Coroutine: Performing ExecutionController.single_step()")
                    # Pick the currently selected thread and step it forward by one
                    # instruction.
                    #
                    # LLDB lets us step any thread that we choose, so, maybe we
                    # should consider letting the caller pick which thread they want
                    # the step to happen in?
                    thread = self.process.GetSelectedThread()
                    assert thread is not None, "Tried to single step, but no thread is selected?"

                    e = lldb.SBError()
                    thread.StepInstruction(False, e)
                    if not e.success:
                        # The step failed. Raise an error in the coroutine and give
                        # it a chance to recover gracefully before we propagate it
                        # up to the caller.
                        exceptions.append(
                            pwndbg.dbg_mod.Error(f"Could not perform single step: {e.description}")
                        )
                        continue

                    status = self._run_until_next_stop()
                    if isinstance(status, _PollResultExited):
                        # The process exited. Cancel the execution controller.
                        exceptions.append(CancelledError())
                        continue

                elif isinstance(step, YieldContinue):
                    self.debug_print("Coroutine: Performing ExecutionController.cont()")
                    threads_suspended = []
                    if step.selected_thread:
                        thread = self.process.GetSelectedThread()
                        for t in self.process.threads:
                            if t.id == thread.id:
                                continue
                            if not t.is_suspended:
                                t.Suspend()
                                threads_suspended.append(t)

                    # Continue the process and wait for the next stop-like event.
                    self.process.Continue()
                    status = self._run_until_next_stop()

                    if threads_suspended:
                        for t in threads_suspended:
                            if t.IsValid():
                                t.Resume()

                    match status:
                        case _PollResultExited():
                            # The process exited, Cancel the execution controller.
                            exceptions.append(CancelledError())
                            continue
                        case _PollResultStopped(event):
                            pass
                        case _:
                            raise AssertionError(f"unexpected poll result {status}")

                    # Check whether this stop event is the one we expect.
                    stop: lldb.SBBreakpoint | lldb.SBWatchpoint = step.target.inner

                    if lldb.SBProcess.GetStateFromEvent(event) == lldb.eStateStopped:
                        matches = 0
                        for thread in lldb.SBProcess.GetProcessFromEvent(event).threads:
                            # We only check the stop reason, as the other methods
                            # for querying thread state (`IsStopped`, `IsSuspended`)
                            # are unreliable[1][2], and so we just assume that
                            # after a stop event, all the threads are stopped[3].
                            #
                            # [1]: https://github.com/llvm/llvm-project/issues/16196
                            # [2]: https://discourse.llvm.org/t/bug-28455-new-thread-state-not-in-sync-with-process-state/41699
                            # [3]: https://discourse.llvm.org/t/sbthread-isstopped-always-returns-false-on-linux/36944/5

                            bpwp_id = None
                            if thread.GetStopReason() == lldb.eStopReasonBreakpoint and isinstance(
                                stop, lldb.SBBreakpoint
                            ):
                                bpwp_id = thread.GetStopReasonDataAtIndex(0)
                            elif (
                                thread.GetStopReason() == lldb.eStopReasonWatchpoint
                                and isinstance(stop, lldb.SBWatchpoint)
                            ):
                                bpwp_id = thread.GetStopReasonDataAtIndex(0)

                            if bpwp_id is not None and stop.GetID() == bpwp_id:
                                matches += 1

                        if matches > 0:
                            # At least one of the threads got stopped by our target.
                            # Return control back to the coroutine and await further
                            # instruction.
                            pass
                        else:
                            # Something else that we weren't expecting caused the
                            # process to stop. Request that the coroutine be
                            # cancelled.
                            exceptions.append(CancelledError())
                    else:
                        # The process might've crashed, been terminated, exited, or
                        # we might've lost connection to it for some other reason.
                        # Regardless, we should cancel the coroutine.
                        exceptions.append(CancelledError())

        # Let the caller distinguish between a coroutine that's been run to
        # completion and one that got cancelled.
        if len(exceptions) > 0:
            return not isinstance(exceptions[-1], CancelledError)
        return True

    def _prepare_listener_for(self, target: lldb.SBTarget):
        """
        Prepares the internal event listener for the given target.
        """
        self.listener = lldb.SBListener("pwndbg.dbg_mod.lldb.repl.proc.ProcessDriver")
        assert self.listener.IsValid()

        self.listener.StartListeningForEvents(
            target.broadcaster,
            lldb.SBTarget.eBroadcastBitModulesLoaded,
        )

    def _enter(self, enter: Callable[..., lldb.SBError], *args) -> LaunchResult:
        """
        Internal logic helper for launch and attach.

        Assumes `self.listener` is correctly initialized, and that after `enter`
        returns, `self.process` is correctly initialized and in a stopped state
        and `self.io` is correctly initialized.
        """
        assert not self.has_process(), "called _enter() on a driver with a live process"

        # Do the launch, proper. We always stop the target, and let the upper
        # layers deal with the user wanting the program to not stop at entry by
        # calling `cont()`.
        error = enter(*args)

        if not error.success:
            # Undo initialization or drop the connection.
            #
            # Ideally with remote targets we would at least keep the connection,
            # but LLDB is rather frail when it comes to preverving it gracefully
            # across failures, so we always drop everything.
            self.process = None
            self.listener = None

            return LaunchResultError(error, disconnected=False)

        assert self.listener.IsValid()
        assert self.process.IsValid()

        result = self._run_until_next_stop(fire_events=False)
        match result:
            case _PollResultExited():
                return LaunchResultEarlyExit()
            case _PollResultStopped():
                pass
            case _:
                raise AssertionError(f"unexpected poll result {type(result)}")

        self.eh.created()

        return LaunchResultSuccess()

    def _launch_remote(
        self,
        env: list[str],
        args: list[str],
        working_dir: str | None,
        extra_flags: int,
    ) -> lldb.SBError:
        """
        Launch a process in a remote debugserver.

        This function always uses a plain text IODriver, as there is no way to
        guarantee any other driver will work.
        """
        assert self.io is None
        self.io = IODriverPlainText()

        error = lldb.SBError()
        stdin, stdout, stderr = self.io.stdio()
        self.process.RemoteLaunch(
            args,
            env,
            stdin,
            stdout,
            stderr,
            working_dir,
            lldb.eLaunchFlagStopAtEntry | extra_flags,
            True,
            error,
        )
        return error

    def _launch_local(
        self,
        target: lldb.SBTarget,
        io: IODriver,
        env: list[str],
        args: list[str],
        working_dir: str | None,
        extra_flags: int,
    ) -> lldb.SBError:
        """
        Launch a process in the host system.
        """
        assert self.io is None
        self.io = io

        error = lldb.SBError()
        stdin, stdout, stderr = io.stdio()
        self.process = target.Launch(
            self.listener,
            args,
            env,
            stdin,
            stdout,
            stderr,
            working_dir,
            lldb.eLaunchFlagStopAtEntry | extra_flags,
            True,
            error,
        )
        return error

    def _attach_remote(self, pid: int) -> lldb.SBError:
        """
        Attach to a process in a remote debugserver.
        """
        if pid == 0:
            return lldb.SBError("PID of 0 or no PID was given")

        assert self.io is None
        self.io = IODriverPlainText()

        error = lldb.SBError()
        self.process.RemoteAttachToProcessWithID(pid, error)
        return error

    def _attach_local(self, target: lldb.SBTarget, info: lldb.SBAttachInfo) -> lldb.SBError:
        """
        Attatch to a process in the host system.
        """
        assert self.io is None
        self.io = IODriverPlainText()

        error = lldb.SBError()
        info.SetListener(self.listener)
        self.process = target.Attach(info, error)
        return error

    def launch(
        self,
        target: lldb.SBTarget,
        io: IODriver,
        env: list[str],
        args: list[str],
        working_dir: str | None,
        disable_aslr: bool,
    ) -> LaunchResult:
        """
        Launches the process and handles startup events. Always stops on first
        opportunity, and returns immediately after the process has stopped.

        Fires the created() event.
        """
        extra_flags = 0
        if disable_aslr:
            extra_flags |= lldb.eLaunchFlagDisableASLR

        if self.has_connection():
            result = self._enter(self._launch_remote, env, args, working_dir, extra_flags)
            if isinstance(result, LaunchResultError):
                result.disconnected = True
            return result
        self._prepare_listener_for(target)
        return self._enter(self._launch_local, target, io, env, args, working_dir, extra_flags)

    def launch_core_file(self, target: lldb.SBTarget, core: str) -> LaunchResult:
        """
        Launches the given core file and puts the process driver into core file
        mode.

        In core file mode, using lifetime-related functionality - continuing,
        running coroutines, etc. - is forbidden. Core file mode ends when the
        process exits.

        Fires the created() and suspended() events.
        """

        # The _enter machinery runs the event loop as part of initializing the
        # new process. We don't want that here. Processes built by LoadCore
        # don't broadcast lifetime events, and using _run_until_next_stop when
        # there's nothing to listen to is iffy at best.

        error = lldb.SBError()
        self.process = target.LoadCore(core, error)

        if not error.success:
            self.process = None
            return LaunchResultError(error, False)

        assert self.process.IsValid()
        self.debug_print("Core File: Loaded")
        self.debug_print("Core File: All events will be simulated by ProcessDriver")

        # Enter core file mode, disabling all lifetime controls in the process
        # driver, as well as the event loop. This ensures that the driver won't
        # get stuck trying to either control or handle events from a process
        # that does not support it.
        self._is_core_file = True

        # Still fire the created event as normal. Process objects based on core
        # files are created in the stopped state, and stay in it for the
        # duration of their lifetimes.
        #
        # From the point of view of the rest of pwndbg, this is no different
        # from firing the event after having run _wait_until_next_stop.
        self.debug_print("Core File: created()")
        self.eh.created()
        self.debug_print("Core File: suspended(None)")
        self.eh.suspended(None)

        return LaunchResultSuccess()

    def attach(self, target: lldb.SBTarget, info: lldb.SBAttachInfo) -> LaunchResult:
        """
        Attach to a process and handles startup events. Always stops on first
        opportunity, and returns immediately after the process has stopped.

        Fires the created() event.
        """
        if self.has_connection():
            result = self._enter(self._attach_remote, info.GetProcessID())
            if isinstance(result, LaunchResultError):
                result.disconnected = True
            return result
        self._prepare_listener_for(target)
        return self._enter(self._attach_local, target, info)

    def connect(self, target: lldb.SBTarget, io: IODriver, url: str, plugin: str) -> LaunchResult:
        """
        Connects to a remote proces with the given URL using the plugin with the
        given name. This might cause the process to launch in some implementations,
        or it might require a call to `launch()`, in implementations that
        require a further call to `SBProcess::RemoteLaunch()`.

        Fires the created() event if a process is automatically attached to
        or launched when a connection succeeds.
        """
        assert not self.has_connection(), "called connect() on a driver with an active connection"
        stdin, stdout, stderr = io.stdio()
        error = lldb.SBError()

        self._prepare_listener_for(target)

        # Connect to the given remote URL using the given remote process plugin.
        self.process = target.ConnectRemote(self.listener, url, plugin, error)

        if not error.success:
            # Undo any initialization ConnectRemote might've done.
            self.process = None
            self.listener = None
            return LaunchResultError(error, False)

        assert self.listener.IsValid()
        assert self.process.IsValid()

        assert self.io is None
        self.io = io

        # It's not guaranteed that the process is actually alive, as it might be
        # in the Connected state, which indicates that the connection was
        # successful, but that we still need to launch it manually in the remote
        # target.
        while True:
            result = self._run_until_next_stop(
                with_io=False, fire_events=False, only_if_started=True
            )
            match result:
                case _PollResultStopped():
                    # The process has started. We can fire off the created event
                    # just fine.
                    self.eh.created()
                    return LaunchResultSuccess()
                case _PollResultExited():
                    # The process quit before we could do anything.
                    return LaunchResultEarlyExit()
                case _PollResultTimedOut(last_event):
                    # Timed out.
                    if (
                        last_event is not None
                        and lldb.SBProcess.GetStateFromEvent(last_event) == lldb.eStateConnected
                    ):
                        # This indicates that on this implementation, connecting isn't
                        # enough to have the process launch or attach, and that we have
                        # to do that manually.
                        return LaunchResultConnected()
                case _:
                    raise AssertionError(f"unexpected poll result {type(result)}")



"""
Helpers for finding address mappings which are used as a stack.

Generally not needed, except under qemu-user and for when
binaries do things to remap the stack (e.g. pwnies' postit).
"""

from __future__ import annotations

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.aglib.vmmap_custom
import pwndbg.color.message as message
import pwndbg.lib.cache
import pwndbg.lib.config
import pwndbg.lib.memory

auto_explore = pwndbg.config.add_param(
    "auto-explore-stack",
    "warn",
    "stack exploration; it may be really slow",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["warn", "yes", "no"],
)


def find(address: int) -> pwndbg.lib.memory.Page | None:
    """
    Returns a pwndbg.lib.memory.Page object which corresponds to given address stack
    or None if it does not exist
    """
    for stack in get().values():
        if address in stack:
            return stack

    return None


def find_upper_stack_boundary(stack_ptr: int, max_pages: int = 1024) -> int:
    stack_ptr = pwndbg.lib.memory.page_align(int(stack_ptr))

    # We can't get the stack size from stack layout and page fault on bare metal mode,
    # so we return current page as a walkaround.
    if not pwndbg.aglib.memory.is_pagefault_supported():
        return stack_ptr + pwndbg.aglib.memory.PAGE_SIZE

    return pwndbg.aglib.memory.find_upper_boundary(stack_ptr, max_pages)


@pwndbg.lib.cache.cache_until("stop")
def get() -> dict[int, pwndbg.lib.memory.Page]:
    """
    For each running thread, return the known address range for its stack
    Returns a dict which should never be modified (since its cached)
    """
    stacks = _fetch_via_vmmap()

    if stacks:
        return stacks

    # Note: exploration is slow
    return _fetch_via_exploration()


@pwndbg.lib.cache.cache_until("stop")
def current() -> pwndbg.lib.memory.Page | None:
    """
    Returns the bounds for the stack for the current thread.
    """
    return find(pwndbg.aglib.regs.sp)


@pwndbg.lib.cache.cache_until("start")
def is_executable() -> bool:
    ehdr = pwndbg.aglib.elf.exe()
    if ehdr is None:
        return True

    for phdr in pwndbg.aglib.elf.iter_phdrs(ehdr):
        # check if type is PT_GNU_STACK
        if phdr.p_type == 0x6474E551:
            return False

    return True


def _fetch_via_vmmap() -> dict[int, pwndbg.lib.memory.Page]:
    stacks: dict[int, pwndbg.lib.memory.Page] = {}

    pages = pwndbg.aglib.vmmap.get()

    for thread in pwndbg.dbg.selected_inferior().threads():
        with thread.bottom_frame() as frame:
            sp = frame.sp()

        # Skip if sp is 0 (it might be 0 if we debug a qemu kernel)
        if not sp:
            continue

        # Find the given SP in pages
        page = next((page for page in pages if sp in page), None)
        if not page:
            # TODO: Handle case where the page is not found;
            #  consider exploring the `sp` register using method `_fetch_via_exploration`?
            continue
        stacks[thread.index()] = page

    return stacks


def _fetch_via_exploration() -> dict[int, pwndbg.lib.memory.Page]:
    """
    TODO/FIXME: This exploration is not great since it now hits on each stop
    (based on how this function is used). Ideally, explored stacks should be
    cached globally and cleared only with new debugged target.

    This way, we should also explore the stack only for a maximum of few pages
    so that we won't take too much time finding its bounds. Then, on each stop
    we can explore one more (or a few more) pages for the given current stack
    we are currently on, ideally not taking the precious time of our users.

    An alternative to this is dumping this functionality completely and this
    will be decided hopefully after a next release.
    """
    if auto_explore.value == "warn":
        print(
            message.warn(
                "Warning: All methods to detect STACK have failed.\n"
                "You can explore STACK using exploration, but it may be very slow.\n"
                "To explicitly explore, use the command: `stack-explore`\n"
                "Alternatively, enable it by default with: `set auto-explore-stack yes`"
            )
        )
        return {}
    if auto_explore.value == "no":
        return {}

    thread_sp = []
    for thread in pwndbg.dbg.selected_inferior().threads():
        with thread.bottom_frame() as frame:
            sp = frame.sp()

        # Skip if sp is None or 0
        # (it might be 0 if we debug a qemu kernel)
        if not sp:
            continue
        thread_sp.append((sp, thread.index()))

    # Sort by the `sp` register (stack pointer), starting with the smallest value.
    # This helps prevent scanning the same page multiple times.
    thread_sp.sort(key=lambda t: t[0])

    ptrsize: int = pwndbg.aglib.arch.ptrsize
    stacks: dict[int, pwndbg.lib.memory.Page] = {}
    for sp, thread_idx in thread_sp:
        start = pwndbg.lib.memory.page_align(sp) - pwndbg.lib.memory.PAGE_SIZE

        page_found = next((page for page in stacks.values() if start in page), None)
        if page_found:
            # Skip further exploration of stacks that have already been scanned.
            stacks[thread_idx] = page_found
            continue

        stop = find_upper_stack_boundary(sp)
        page = pwndbg.lib.memory.Page(
            start,
            stop - start,
            6 if not is_executable() else 7,
            0,
            ptrsize,
            f"[stack:{thread_idx}]",
        )
        stacks[thread_idx] = page
        pwndbg.aglib.vmmap_custom.add_custom_page(page)

    return stacks


def callstack() -> list[int]:
    """
    Return the address of the return address for the current frame.
    """
    frame = pwndbg.dbg.selected_frame()
    addresses = []
    while frame:
        addr = frame.pc()
        if pwndbg.aglib.memory.is_readable_address(addr):
            addresses.append(addr)
        frame = frame.parent()

    return addresses


@pwndbg.lib.cache.cache_until("stop", "start")
def get_stack_var_name(address: int) -> str | None:
    """
    Get the name of the stack variable covering the given address.

    Returns the variable name with optional offset (e.g., "buf+0x8") if the address
    is within a variable but not at its start. Returns None if no variable is found.
    """
    frame = pwndbg.dbg.selected_frame()
    if frame is None:
        return None

    for start, end, name in frame.stack_variables():
        if start <= address < end:
            offset = address - start
            return name if offset == 0 else f"{name}+{offset:#x}"

    return None



from __future__ import annotations

from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.riscv import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib
import pwndbg.aglib.disasm.arch
import pwndbg.color.memory as mem_color
import pwndbg.dintegration
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.arch import register_assign
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

RISCV_LOAD_INSTRUCTIONS = {
    # Sign-extend loads
    RISCV_INS_LB: -1,
    RISCV_INS_LH: -2,
    RISCV_INS_LW: -4,
    # Zero-extend loads
    RISCV_INS_LBU: 1,
    RISCV_INS_LHU: 2,
    RISCV_INS_LWU: 4,
    RISCV_INS_LD: 8,
    RISCV_INS_C_LW: -4,
    RISCV_INS_C_LWSP: -4,
    RISCV_INS_C_LD: 8,
    RISCV_INS_C_LDSP: 8,
}

RISCV_STORE_INSTRUCTIONS = {
    RISCV_INS_SB: 1,
    RISCV_INS_SH: 2,
    RISCV_INS_SW: 4,
    RISCV_INS_SD: 8,
    RISCV_INS_C_SW: 4,
    RISCV_INS_C_SWSP: 4,
    RISCV_INS_C_SD: 8,
    RISCV_INS_C_SDSP: 8,
}


RISCV_MATH_INSTRUCTIONS = {
    RISCV_INS_ADDI: "+",
    RISCV_INS_ADD: "+",
    RISCV_INS_C_ADDI: "+",
    RISCV_INS_C_ADD: "+",
    RISCV_INS_SUB: "-",
    RISCV_INS_C_SUB: "-",
    RISCV_INS_XORI: "^",
    RISCV_INS_XOR: "^",
    RISCV_INS_C_XOR: "^",
    RISCV_INS_ORI: "|",
    RISCV_INS_OR: "|",
    RISCV_INS_C_OR: "|",
    RISCV_INS_ANDI: "&",
    RISCV_INS_C_ANDI: "&",
    RISCV_INS_AND: "&",
    RISCV_INS_C_AND: "&",
    RISCV_INS_SLLI: "<<",
    RISCV_INS_C_SLLI: "<<",
    RISCV_INS_SLL: "<<",
    RISCV_INS_SRLI: ">>",
    RISCV_INS_C_SRLI: ">>",
    RISCV_INS_SRL: ">>",
    RISCV_INS_SRAI: ">>s",
    RISCV_INS_C_SRAI: ">>s",
    RISCV_INS_SRA: ">>s",
    RISCV_INS_MUL: "*",
    RISCV_INS_MULH: "*",
    RISCV_INS_MULHSU: "*",
    RISCV_INS_MULHU: "*",
    RISCV_INS_DIV: "/",
    RISCV_INS_DIVU: "/",
    RISCV_INS_REM: "%",
    RISCV_INS_REMU: "%",
    RISCV_INS_C_ADDI4SPN: "+",
    RISCV_INS_C_ADDI16SP: "+",
    # RV64I unique instructions
    RISCV_INS_ADDIW: "+",
    RISCV_INS_ADDW: "+",
    RISCV_INS_SUBW: "-",
    RISCV_INS_SLLIW: "<<",
    RISCV_INS_SLLW: "<<",
    RISCV_INS_SRLIW: ">>",
    RISCV_INS_SRLW: ">>",
    RISCV_INS_SRAIW: ">>s",
    RISCV_INS_SRAW: ">>s",
    # RV64M unique instructions
    RISCV_INS_MULW: "*",
    RISCV_INS_DIVW: "/",
    RISCV_INS_DIVUW: "/",
    RISCV_INS_REMW: "%",
    RISCV_INS_REMUW: "%",
    # RV64C unique instructions
    RISCV_INS_C_ADDIW: "+",
    RISCV_INS_C_ADDW: "+",
    RISCV_INS_C_SUBW: "-",
}

RISCV_EMULATED_ANNOTATIONS = {
    RISCV_INS_SLT,
    RISCV_INS_SLTU,
    RISCV_INS_SLTI,
    RISCV_INS_SLTIU,
}


class RISCVDisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    def __init__(self, architecture) -> None:
        super().__init__(architecture)
        self.architecture = architecture

        self.annotation_handlers: Dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {
            # AUIPC
            RISCV_INS_AUIPC: self._auipc_annotator,
            # C.MV
            RISCV_INS_C_MV: self._common_move_annotator,
            # C.LI
            RISCV_INS_C_LI: self._common_move_annotator,
            # LUI
            RISCV_INS_LUI: self._lui_annotator,
            RISCV_INS_C_LUI: self._lui_annotator,
        }

    @override
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        if instruction.id in RISCV_LOAD_INSTRUCTIONS:
            read_size = RISCV_LOAD_INSTRUCTIONS[instruction.id]
            self._common_load_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                abs(read_size),
                read_size < 0,
                pwndbg.aglib.arch.ptrsize,
                instruction.operands[0].str,
                instruction.operands[1].str,
            )
        elif instruction.id in RISCV_STORE_INSTRUCTIONS:
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                instruction.operands[0].before_value,
                RISCV_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[1].str,
            )
        elif instruction.id in RISCV_MATH_INSTRUCTIONS:
            # We need this check, because some of these instructions can encoded as aliases
            # Example: NOP is an alias of ADDI where target is x0. In Capstone, the ID will still be that of ADDI but with no operands
            if len(instruction.operands) >= 2:
                self._common_binary_op_annotator(
                    instruction,
                    emu,
                    instruction.operands[0],
                    instruction.operands[-2].before_value,
                    instruction.operands[-1].before_value,
                    RISCV_MATH_INSTRUCTIONS[instruction.id],
                )
        elif instruction.id in RISCV_EMULATED_ANNOTATIONS:
            self._common_generic_register_destination(instruction, emu)
        else:
            self.annotation_handlers.get(instruction.id, lambda *a: None)(instruction, emu)

    def _auipc_annotator(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        result_operand, right = instruction.operands
        if result_operand.str and right.before_value is not None:
            if (address := result_operand.after_value) is None:
                # Resolve it manually without emulation
                address = instruction.address + (right.before_value << 12)

            instruction.annotation = register_assign(
                result_operand.str,
                mem_color.get_address_and_symbol(
                    address, pwndbg.dintegration.manager.get_stack_var_dict_all()
                ),
            )

    def _lui_annotator(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        result_operand, right = instruction.operands
        if result_operand.str and right.before_value is not None:
            if (address := result_operand.after_value) is None:
                # Resolve it manually without emulation
                address = right.before_value << 12

            instruction.annotation = register_assign(
                result_operand.str,
                mem_color.get_address_and_symbol(
                    address, pwndbg.dintegration.manager.get_stack_var_dict_all()
                ),
            )

    def _is_condition_taken(
        self, instruction: PwndbgInstruction, emu: Emulator | None
    ) -> InstructionCondition:
        # B-type instructions have two source registers that are compared
        src1_unsigned = instruction.op_find(CS_OP_REG, 1).before_value
        # compressed instructions c.beqz and c.bnez only use one register operand.
        if instruction.op_count(CS_OP_REG) > 1:
            src2_unsigned = instruction.op_find(CS_OP_REG, 2).before_value
        else:
            src2_unsigned = 0

        if src1_unsigned is None or src2_unsigned is None:
            return InstructionCondition.UNDETERMINED

        src1_signed = bit_math.to_signed(src1_unsigned, pwndbg.aglib.arch.ptrbits)
        src2_signed = bit_math.to_signed(src2_unsigned, pwndbg.aglib.arch.ptrbits)

        condition = {
            RISCV_INS_BEQ: src1_signed == src2_signed,
            RISCV_INS_BNE: src1_signed != src2_signed,
            RISCV_INS_BLT: src1_signed < src2_signed,
            RISCV_INS_BGE: src1_signed >= src2_signed,
            RISCV_INS_BLTU: src1_unsigned < src2_unsigned,
            RISCV_INS_BGEU: src1_unsigned >= src2_unsigned,
            RISCV_INS_C_BEQZ: src1_signed == 0,
            RISCV_INS_C_BNEZ: src1_signed != 0,
        }.get(instruction.id, None)

        if condition is None:
            return InstructionCondition.UNDETERMINED

        return InstructionCondition.TRUE if bool(condition) else InstructionCondition.FALSE

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        """
        Checks if the current instruction is a jump that is taken.
        """
        # JAL / JALR is unconditional
        if RISCV_GRP_CALL in instruction.groups:
            return InstructionCondition.UNDETERMINED

        # Determine if the conditional jump is taken
        if RISCV_GRP_BRANCH_RELATIVE in instruction.groups:
            return self._is_condition_taken(instruction, emu)

        return InstructionCondition.UNDETERMINED

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        """Return the address of the jump / conditional jump,
        None if the next address is not dependent on instruction.
        """
        ptrmask = pwndbg.aglib.arch.ptrmask
        # JAL is unconditional and independent of current register status
        if instruction.id in (RISCV_INS_JAL, RISCV_INS_C_JAL, RISCV_INS_C_J):
            # But that doesn't apply to ARM anyways :)
            return (instruction.address + instruction.op_find(CS_OP_IMM, 1).imm) & ptrmask

        # Determine target of branch - all of them are offset to address
        if RISCV_GRP_BRANCH_RELATIVE in instruction.groups:
            return (instruction.address + instruction.op_find(CS_OP_IMM, 1).imm) & ptrmask

        # Determine the target address of the indirect jump
        if instruction.id == RISCV_INS_JALR:
            # jalr can be represented in the following ways:
            # 1. jalr rd                // Jump to rd
            # 2. jalr rd, offset        // Jump to rd+offset
            # 3. jalr rX, rd, offset    // Return address stored in rX, jump to rd+offset
            # 4. jalr x0, x1, 0         // Disassembles as "ret", jump to ra

            # To find target, get the LAST register
            reg_op_count = instruction.op_count(CS_OP_REG)

            if reg_op_count == 0:
                # ra is implied as link register
                return self._read_register_name(instruction, "ra", emu)

            if (target := instruction.op_find(CS_OP_REG, reg_op_count).before_value) is None:
                return None

            if (imm_op := instruction.op_find(CS_OP_IMM, 1)) is not None:
                target += imm_op.imm
            target &= ptrmask
            # Clear the lowest bit without knowing the register width
            return target ^ (target & 1)

        if instruction.id == RISCV_INS_C_JALR:
            if (target := instruction.op_find(CS_OP_REG, 1).before_value) is None:
                return None
            return target ^ (target & 1)

        return super()._resolve_target(instruction, emu)

    @override
    def _parse_memory(
        self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator
    ) -> int | None:
        """
        Parse the `RISCVOpMem` Capstone object to determine the concrete memory address used.
        """
        base = self._read_register(instruction, op.mem.base, emu)
        if base is None:
            return None
        return base + op.mem.disp



"""
Answers libc-specific questions in a non-commital way when
we don't know which libc implementation we are actually using.
"""

from __future__ import annotations

from .dispatch import LibcType
from .dispatch import LibcURLs


def type() -> LibcType:
    return LibcType.UNKNOWN


def _is_being_used() -> bool:
    return True


def version(libc_filepath: str) -> tuple[int, ...]:
    return (-1, -1)


def has_internal_symbols(libc_filepath: str) -> bool:
    return False


def has_debug_info() -> bool:
    return False


def verify_libc_candidate(mapping_name: str) -> bool:
    # We cheat a bit and return False for both
    # verify_libc_candidate and verify_ld_candidate.
    return False


def verify_ld_candidate(mapping_name: str) -> bool:
    return False


def urls(ver: tuple[int, ...] | None) -> LibcURLs:
    return LibcURLs(
        versioned_readable_source="not available",
        versioned_compressed_source="not available",
        homepage="not available",
        git="not available",
    )


def libc_same_as_ld() -> bool:
    return False



"""
Functionality for resolving ASCII printable strings within
the debuggee's address space.
"""

from __future__ import annotations

import re
import string
from collections.abc import Iterator

import pwndbg
import pwndbg.aglib.memory
import pwndbg.dbg_mod
from pwndbg.lib.memory import Page

length = 15


def update_length() -> None:
    r"""
    Unfortunately there's not a better way to get at this info.

    >>> gdb.execute('show print elements', from_tty=False, to_string=True)
    'Limit on string chars or array elements to print is 21.\n'
    """
    global length
    length = pwndbg.dbg.string_limit()


def get(address: int, maxlen: int | None = None, maxread: int | None = None) -> str | None:
    """
    Returns a printable C-string from address.

    Returns `None` if string contains non-printable chars
    or if the `maxlen` length data does not end up with a null byte.
    """
    if maxlen is None:
        maxlen = length

    if maxread is None:
        maxread = length

    try:
        bytesz = pwndbg.aglib.memory.string(address, maxread)
    except pwndbg.dbg_mod.Error:  # should not happen, but sanity check?
        return None

    sz = bytesz.decode("latin-1", "replace")

    if not sz or not all(s in string.printable for s in sz):
        return None

    if len(sz) < maxlen or not maxlen:
        return sz

    return sz[:maxlen] + "..."


def yield_in_page(page: Page, n=4) -> Iterator[str]:
    """Yields strings of length >= n found in a given vmmap page"""
    try:
        data = pwndbg.aglib.memory.read(addr=page.vaddr, count=page.memsz, partial=True)
    except pwndbg.dbg_mod.Error:
        # E.g. we cannot read [vvar] page even though it has a READ permission
        return

    for match in re.finditer(rb"[ -~]{%d,}" % n, data):
        decoded_str = match.group().decode("ascii", errors="ignore")
        yield decoded_str



from __future__ import annotations

import argparse
import contextlib
import socket
from typing import Literal
from typing import NamedTuple
from urllib.parse import ParseResult
from urllib.parse import urlparse

from pwnlib import constants
from pwnlib import shellcraft
from pwnlib.util.net import sockaddr

import pwndbg.aglib.asm
import pwndbg.aglib.memory
import pwndbg.aglib.shellcode
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib.regs
from pwndbg.commands import CommandCategory


class ShellcodeRegs(NamedTuple):
    newfd: str
    syscall_ret: str
    stack: str


def get_shellcode_regs() -> ShellcodeRegs:
    register_set = pwndbg.lib.regs.reg_sets[pwndbg.aglib.arch.name]
    syscall_abi = pwndbg.aglib.arch.syscall_abi
    if syscall_abi is None:
        raise pwndbg.dbg_mod.Error("Syscall ABI not defined for current architecture")

    # pickup free register what is not used for syscall abi
    newfd_reg = next(
        reg_name
        for reg_name in register_set.gpr
        if reg_name not in syscall_abi.register_arguments
        and reg_name != syscall_abi.syscall_register
    )
    assert newfd_reg is not None, (
        f"architecture {pwndbg.aglib.arch.name} don't have unused register..."
    )

    return ShellcodeRegs(newfd_reg, register_set.retval, register_set.stack)


def stack_size_alignment(s: int) -> int:
    syscall_abi = pwndbg.aglib.arch.syscall_abi
    if syscall_abi is None:
        raise pwndbg.dbg_mod.Error("Syscall ABI not defined for current architecture")

    return s + (syscall_abi.arg_alignment - (s % syscall_abi.arg_alignment))


def asm_replace_file(replace_fd: int, filename: str) -> tuple[int, bytes]:
    filename = filename.encode() + b"\x00"

    regs = get_shellcode_regs()
    stack_size = stack_size_alignment(len(filename))

    open_asm = (
        shellcraft.syscall("SYS_open", regs.stack, "O_CREAT|O_RDWR", 0o666)
        if hasattr(constants, "SYS_open")
        else shellcraft.syscall("SYS_openat", "AT_FDCWD", regs.stack, "O_CREAT|O_RDWR", 0o666)
    )

    dup_asm = (
        shellcraft.syscall("SYS_dup2", regs.newfd, replace_fd)
        if hasattr(constants, "SYS_dup2")
        else shellcraft.syscall("SYS_dup3", regs.newfd, replace_fd, 0)
    )

    return stack_size, pwndbg.aglib.asm.asm(
        "".join(
            [
                shellcraft.pushstr(filename, False),
                open_asm,
                shellcraft.mov(regs.newfd, regs.syscall_ret),
                dup_asm,
                shellcraft.syscall("SYS_close", regs.newfd),
            ]
        )
    )


def asm_replace_socket(replace_fd: int, socket_data: ParsedSocket) -> tuple[int, bytes]:
    sockdata, addr_len, _ = sockaddr(socket_data.address, socket_data.port, socket_data.ip_version)
    socktype = {"tcp": "SOCK_STREAM", "udp": "SOCK_DGRAM"}[socket_data.protocol]
    family = {"ipv4": "AF_INET", "ipv6": "AF_INET6"}[socket_data.ip_version]

    regs = get_shellcode_regs()
    stack_size = stack_size_alignment(len(sockdata))

    dup_asm = (
        shellcraft.syscall("SYS_dup2", regs.newfd, replace_fd)
        if hasattr(constants, "SYS_dup2")
        else shellcraft.syscall("SYS_dup3", regs.newfd, replace_fd, 0)
    )

    return stack_size, pwndbg.aglib.asm.asm(
        "".join(
            [
                shellcraft.syscall("SYS_socket", family, socktype, 0),
                shellcraft.mov(regs.newfd, regs.syscall_ret),
                shellcraft.pushstr(sockdata, False),
                shellcraft.syscall("SYS_connect", regs.newfd, regs.stack, addr_len),
                dup_asm,
                shellcraft.syscall("SYS_close", regs.newfd),
            ]
        )
    )


@contextlib.asynccontextmanager
async def exec_shellcode_with_stack(ec: pwndbg.dbg_mod.ExecutionController, blob, stack_size: int):
    # This function could be improved, for example:
    # - Run the shellcode inside an emulator like Unicorn
    # - Calculate the maximum stack size the shellcode would consume dynamically.

    stack_start_diff = pwndbg.aglib.regs.sp
    stack_start = stack_start_diff - stack_size
    original_stack = pwndbg.aglib.memory.read(stack_start, stack_size)

    try:
        async with pwndbg.aglib.shellcode.exec_shellcode(ec, blob):
            stack_diff_size = stack_start_diff - pwndbg.aglib.regs.sp

            # Make sure stack is not corrupted somehow
            assert not (stack_diff_size > stack_size), (
                f"stack is probably corrupted size_current=f{stack_diff_size} size_max_want={stack_size}"
            )

            yield
    finally:
        pwndbg.aglib.memory.write(stack_start, original_stack)


parser = argparse.ArgumentParser(
    description="""Replace a file descriptor of a debugged process.

The new file descriptor can point to:

- a file
- a pipe
- a socket
- a device, etc.
""",
)

parser.add_argument(
    "fdnum",
    help="File descriptor (FD) number to be replaced with the specified new socket or file.",
    type=int,
)


class ParsedSocket(NamedTuple):
    protocol: Literal["tcp", "udp"]
    ip_version: Literal["ipv4", "ipv6"]
    address: str
    port: int


def parse_socket(url: str) -> ParsedSocket:
    if "://" in url:
        # For handling:
        # - `tcp://[::1]:80`
        # - `udp://example.com:80`
        # - `tcp+ipv6://example.com:80`
        parsed = urlparse(url)
    else:
        # For handling:
        # - `127.0.0.1:80`
        parsed = ParseResult("", url, "", "", "", "")

    # Handling eg: `tcp+ipv6://example.com:80`
    scheme_info = parsed.scheme.split("+", 1)

    selected_protocol: Literal["tcp", "udp"] = "tcp"
    selected_ip_protocol: Literal["ipv4", "ipv6"] | None = None
    if parsed.scheme:
        for any_value in scheme_info:
            if any_value in ("tcp", "udp"):
                selected_protocol = any_value
            elif any_value in ("ipv4", "ipv6"):
                selected_ip_protocol = any_value

    domain_or_ip = parsed.hostname
    if not domain_or_ip:
        raise argparse.ArgumentTypeError("Domain or IP is required")

    port = parsed.port
    if not port:
        raise argparse.ArgumentTypeError("Port is required")

    protocol_ordered = (
        ("ipv4", socket.AF_INET),
        ("ipv6", socket.AF_INET6),
    )

    found_ip_protocol: Literal["ipv4", "ipv6"] | None = None
    address_ipv4_or_ipv6: str = ""
    for family_name, family_const in protocol_ordered:
        if selected_ip_protocol and selected_ip_protocol != family_name:
            continue

        try:
            # Resolve the given domain or IP address to its corresponding IP address
            ips = socket.getaddrinfo(domain_or_ip, None, family_const)
        except socket.gaierror:
            # happen when domain not found
            continue

        for _, _, _, _, ip in ips:
            address_ipv4_or_ipv6 = ip[0]
            found_ip_protocol = family_name

        if found_ip_protocol:
            break

    if not address_ipv4_or_ipv6:
        raise argparse.ArgumentTypeError(
            f"Could not resolve {domain_or_ip} to proper {selected_ip_protocol} address"
        )

    if not found_ip_protocol:
        raise argparse.ArgumentTypeError("Protocol only accept: ipv4,ipv6")

    return ParsedSocket(selected_protocol, found_ip_protocol, address_ipv4_or_ipv6, port)


PARSED_FILE_ARG = tuple[ParsedSocket | None, str | None]


def parse_file_or_socket(s: str) -> PARSED_FILE_ARG:
    # is file
    if s.startswith("/") or s.startswith("./"):
        return None, s
    return parse_socket(s), None


parser.add_argument(
    "newfile",
    help="""Specify a file or a socket.

For files, the filename must start with `/` (e.g., `/etc/passwd`).

For sockets, the following formats are allowed:
- `127.0.0.1:80` (default is TCP)
- `tcp://[::1]:80`
- `udp://example.com:80`
- `tcp+ipv6://example.com:80`
    """,
    type=parse_file_or_socket,
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MISC,
    examples="""
1. Redirect STDOUT to a file:
   `hijack-fd 1 /dev/null`

2. Redirect STDERR to a socket:
   `hijack-fd 2 tcp://localhost:8888`
    """,
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def hijack_fd(fdnum: int, newfile: PARSED_FILE_ARG) -> None:
    socket_data, filename = newfile
    if filename:
        stack_size, asm_bin = asm_replace_file(fdnum, filename)
    elif socket_data:
        stack_size, asm_bin = asm_replace_socket(fdnum, socket_data)
    else:
        assert False

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        async with exec_shellcode_with_stack(ec, asm_bin, stack_size):
            print(
                "Operation succeeded. Errors are not captured.\n"
                "You can verify this with `procinfo` if the file descriptor has been replaced."
            )

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)



"""
Heap Tracking

This module implements runtime tracking of the heap, allowing pwndbg to detect
heap related misbehavior coming from an inferior in real time, which lets us
catch UAF bugs, double frees (and more), and report them to the user.

# Approach
The approach used starting with using breakpoints to hook into the following
libc symbols: `malloc`, `free`, `calloc`, and `realloc`. Each hook has a
reference to a shared instance of the `Tracker` class, which is responsible for
handling the tracking of the chunks of memory from the heap.

The tracker keeps two sorted maps of chunks, for freed and in use chunks, keyed
by their base address. Newly allocated chunks are added to the map of in use
chunks right before an allocating call returns, and newly freed chunks are moved
from the map of in use chunks to the map of free ones right before a freeing
call returns. The tracker is also responsible for installing watchpoints for
free chunks when they're added to the free chunk map and deleting them when
their corresponding chunks are removed from the map.

Additionally, because going through the data structures inside of libc to
determine whether a chunk is free or not is, more often than not, a fairly slow
operation, this module will only do so when it determines its view of the chunks
has diverged from the one in libc in a way that would affect behavior. When such
a diffence is detected, this module will rebuild the chunk maps in the range it
determines to have been affected.

Currently, the way it does this is by deleting and querying from libc the new
status of all chunks that overlap the region of a new allocation when it detects
that allocation overlaps chunks it previously considered free.

This approach lets us avoid a lot of the following linked lists that comes with
trying to answer the allocation status of a chunk, by keeping at hand as much
known-good information as possible about them. Keep in mind that, although it is
much faster than going to libc every time we need to know the allocation status
of a chunk, this approach does have drawbacks when it comes to memory usage.

# Compatibility
Currently module assumes the inferior is using GLibc.

There are points along the code in this module where the assumptions it makes
are explicitly documented and checked to be valid for the current inferior, so
that it may be immediately clear to the user that something has gone wrong if
they happen to not be valid. However, be aware that there may be assumptions
that were not made explicit.

"""

from __future__ import annotations

import gdb
from sortedcontainers import SortedDict

import pwndbg.aglib.heap
import pwndbg.aglib.heap.ptmalloc
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.arguments
import pwndbg.color
import pwndbg.dbg_mod
import pwndbg.lib.cache
from pwndbg.color import message
from pwndbg.lib.cache import CacheUntilEvent

LIBC_NAME = "libc.so.6"
MALLOC_NAME = "malloc"
CALLOC_NAME = "calloc"
REALLOC_NAME = "realloc"
FREE_NAME = "free"

last_issue: str | None = None

# Useful to track possbile collision errors.
PRINT_DEBUG = False

PTRS_COLORS = (
    pwndbg.color.red,
    pwndbg.color.green,
    pwndbg.color.yellow,
    pwndbg.color.blue,
    pwndbg.color.purple,
    pwndbg.color.cyan,
    pwndbg.color.light_gray,
    pwndbg.color.gray,
    pwndbg.color.light_red,
    pwndbg.color.light_green,
    pwndbg.color.light_yellow,
    pwndbg.color.light_blue,
    pwndbg.color.light_purple,
    pwndbg.color.light_cyan,
)


def is_enabled() -> bool:
    """
    Whether the heap tracker in enabled.
    """
    global malloc_enter
    global free_enter

    installed = [malloc_enter is not None, free_enter is not None]

    # Make sure we're not in an inconsistent state.
    assert all(installed) == any(installed)

    return any(installed)


def resolve_address(name: str) -> int | None:
    """
    Checks whether a given symbol is available and part of libc, and returns its
    address.
    """
    # If that fails, try to query for it by using the less precise pwndbg API.
    address = pwndbg.aglib.symbol.lookup_symbol_addr(name)
    if not address:
        # Nothing that we can do here.
        return None

    # TODO: dodanie do lookup_symbol, sprawdzanie do jakiego modulu jest przypisany
    # Try to see if this belongs to libc.
    #
    # This check is, frankly, horrifying, but it's one of the few ways we can
    # check what objfile the address we got is coming from*, and it's better to
    # err on the side of caution here and at least attempt to prevent the wrong
    # symbol from being used, than to return a possibly wrong symbol and have
    # the user wonder why on Earth the heap tracker would be hooking to ld.so.
    #
    # *: A better way would be to use gdb.objfile_from_address, but that's only
    # available in relatively recent versions of GDB.
    info = gdb.execute(f"info symbol {address:#x}", to_string=True, from_tty=False)
    info = info.split(" of ")[-1].split("/")[-1]
    if not info or LIBC_NAME not in info:
        print(
            message.warn(
                f'Found "{name}" that does not seem to belong to {LIBC_NAME}. Refusing to use.'
            )
        )
        return None

    return address


class FreeChunkWatchpoint(gdb.Breakpoint):
    def __init__(self, chunk: Chunk, tracker: Tracker) -> None:
        self.chunk = chunk
        self.tracker = tracker

        language = gdb.execute("show language", to_string=True)
        if "rust" in language:
            loc = f"*({chunk.address:#x} as *mut [u8;{chunk.size:#x}])"
        else:
            loc = f"*(char[{chunk.size:#x}]*){chunk.address:#x}"

        super().__init__(loc, type=gdb.BP_WATCHPOINT, internal=True)

    def stop(self):
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        if not in_program_code_stack():
            # Untracked.
            return False

        # malloc() and free() implementations will often modify the data in the
        # payload of a freed chunk, where the watchpoint is insalled. So, we
        # should not flag accesses done as a result of a call to either.
        if self.tracker.is_performing_memory_management():
            # We explicitly allow this operation.
            return False

        msg = f"Possible use-after-free in {self.chunk.size:#x}-byte chunk at address {self.chunk.address:#x}"
        print(f"[!] {msg}")

        global stop_on_error
        if stop_on_error:
            global last_issue
            last_issue = message.error(msg)
        return stop_on_error


class AllocChunkWatchpoint(gdb.Breakpoint):
    def __init__(self, chunk: Chunk) -> None:
        self.chunk = chunk
        super().__init__(f"*(char[{chunk.size:#x}]*){chunk.address:#x}", internal=True)

    def stop(self) -> bool:
        return False


class Chunk:
    def __init__(self, address: int, size: int, requested_size: int, flags: int) -> None:
        self.address = address
        self.size = size
        self.requested_size = requested_size
        self.flags = flags


# GDB doesn't like having its breakpoints deleted during stop handlers, so we
# defer deletion until the next stop event.
DEFERED_DELETE: list[gdb.Breakpoint] = []


@pwndbg.dbg.event_handler(pwndbg.dbg_mod.EventType.STOP)
def _delete_defered():
    for entry in DEFERED_DELETE:
        entry.delete()
    DEFERED_DELETE.clear()


class Tracker:
    def __init__(self) -> None:
        self.free_chunks: SortedDict[int, Chunk] = SortedDict()
        self.alloc_chunks: SortedDict[int, Chunk] = SortedDict()
        self.free_watchpoints: dict[int, FreeChunkWatchpoint] = {}
        self.memory_management_calls: dict[int, bool] = {}
        self.colorized_heap_ptrs: dict[int, str] = {}

    def is_performing_memory_management(self):
        thread = gdb.selected_thread().global_num
        if thread not in self.memory_management_calls:
            return False
        return self.memory_management_calls[thread]

    def enter_memory_management(self, name: str) -> None:
        thread = gdb.selected_thread().global_num

        # We don't support re-entry.
        if thread in self.memory_management_calls:
            assert not self.memory_management_calls[thread], (
                f"in {name}(): re-entrant calls are not supported"
            )

        self.memory_management_calls[thread] = True

    def exit_memory_management(self) -> None:
        thread = gdb.selected_thread().global_num

        # Make sure we're not doing anything wrong.
        if thread in self.memory_management_calls:
            assert self.memory_management_calls[thread], (
                "exit_memory_management_calls assert failed"
            )

        self.memory_management_calls[thread] = False

    def colorize_ptr(self, ptr: int) -> str:
        """
        Returns colored string of the provided pointer/address
        """
        if colored_ptr := self.colorized_heap_ptrs.get(ptr):
            return colored_ptr

        idx = len(self.colorized_heap_ptrs) % len(PTRS_COLORS)
        colored = PTRS_COLORS[idx](f"{ptr:#x}")

        self.colorized_heap_ptrs[ptr] = colored

        return colored

    def malloc(self, chunk: Chunk) -> None:
        # malloc()s may arbitrarily change the structure of freed blocks, to the
        # point our chunk maps may become invalid, so, we update them here if
        # anything looks wrong.
        lo_i = self.free_chunks.bisect_right(chunk.address)
        hi_i = self.free_chunks.bisect_right(chunk.address + chunk.size)
        if lo_i > 0:
            left_chunk = self.free_chunks.peekitem(index=lo_i - 1)[1]
            if left_chunk.address + left_chunk.size >= chunk.address:
                # Include the element to the left in the update.
                lo_i -= 1

        try:
            if lo_i != hi_i:
                # The newly registered chunk overlaps with chunks we had registered
                # previously, which means our libc shuffled some things around and
                # so we need to update our view of the chunks.
                lo_chunk = self.free_chunks.peekitem(index=lo_i)[1]
                hi_chunk = self.free_chunks.peekitem(index=hi_i - 1)[1]

                lo_addr = lo_chunk.address
                hi_addr = hi_chunk.address + hi_chunk.size

                lo_heap = pwndbg.aglib.heap.ptmalloc.Heap(lo_addr)
                hi_heap = pwndbg.aglib.heap.ptmalloc.Heap(hi_addr - 1)
                assert lo_heap.arena is not None and hi_heap.arena is not None, (
                    "malloc assert failed"
                )

                # TODO: Can this ever actually fail in real world use?
                #
                # It shouldn't be possible, the way glibc implements it[0], to have
                # a contiguous range at time t+1 that overlaps with two or more
                # contiguous ranges that at time t belonged to different heaps.
                #
                # glibc doesn't move or resize its heaps, which means the boundaries
                # between them stay fixed, and, since a chunk can only be created
                # from slicing a heap, the heap used to create the chunk at t+1 must
                # be the same as the one used to create the ranges at t that it
                # overlaps with.
                #
                # The question is, if we were to support other implementations, we
                # couldn't take this behavior for granted. Regardless, if we ever
                # do, it's better to fail here if/when this assumption is violated
                # than to let it become a bug.
                #
                # [0]: https://sourceware.org/glibc/wiki/MallocInternals
                assert lo_heap.start == hi_heap.start and lo_heap.end == hi_heap.end, (
                    "malloc assert start failed"
                )

                # Remove all of our old handlers.
                for i in reversed(range(lo_i, hi_i)):
                    addr, ch = self.free_chunks.popitem(index=i)

                    self.free_watchpoints[addr].enabled = False
                    DEFERED_DELETE.append(self.free_watchpoints[addr])
                    del self.free_watchpoints[addr]

                # Add new handlers in their place. We scan over all of the chunks in
                # the heap in the range of affected chunks, and add the ones that
                # are free.
                allocator = pwndbg.aglib.heap.current
                assert isinstance(allocator, pwndbg.aglib.heap.ptmalloc.GlibcMemoryAllocator), (
                    "malloc allocator assert failed"
                )
                bins_list = [
                    allocator.fastbins(lo_heap.arena.address),
                    allocator.smallbins(lo_heap.arena.address),
                    allocator.largebins(lo_heap.arena.address),
                    allocator.unsortedbin(lo_heap.arena.address),
                ]
                if allocator.has_tcache():
                    bins_list.append(allocator.tcachebins(None))
                bins_list = [x for x in bins_list if x is not None]

                for ch in lo_heap:
                    # Check for range overlap.
                    ch_lo_addr = ch.address
                    ch_hi_addr = ch.address + ch.size
                    ch.address

                for ch in lo_heap:
                    # Check for range overlap.
                    ch_lo_addr = ch.address
                    ch_hi_addr = ch.address + ch.size

                    lo_in_range = ch_lo_addr < hi_addr
                    hi_in_range = ch_hi_addr > lo_addr

                    if not lo_in_range or not hi_in_range:
                        # No range overlap.
                        continue

                    # Check if the chunk is free.
                    for b in bins_list:
                        if b.contains_chunk(ch.real_size, ch.address):
                            # The chunk is free. Add it to the free list and install
                            # a new watch point for it.
                            nch = Chunk(ch.address, ch.size, ch.real_size, 0)
                            wp = FreeChunkWatchpoint(nch, self)

                            self.free_chunks[ch.address] = nch
                            self.free_watchpoints[ch.address] = wp

                            # Move on to the next chunk.
                            break
        except IndexError:
            import traceback

            traceback.print_exc()

        self.alloc_chunks[chunk.address] = chunk

    def free(self, address: int) -> bool:
        if address not in self.alloc_chunks:
            return False
        chunk = self.alloc_chunks[address]
        del self.alloc_chunks[address]

        wp = FreeChunkWatchpoint(chunk, self)

        self.free_chunks[chunk.address] = chunk
        self.free_watchpoints[chunk.address] = wp

        return True


class MallocEnterBreakpoint(gdb.Breakpoint):
    def __init__(self, address, tracker) -> None:
        super().__init__(f"*{address:#x}", internal=True)
        self.tracker = tracker

    def stop(self) -> bool:
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        requested_size = pwndbg.arguments.argument(0)
        if self.tracker.is_performing_memory_management():
            # This call was made from inside another memory management call.
            # Ignore it.
            return False

        self.tracker.enter_memory_management(MALLOC_NAME)
        AllocExitBreakpoint(self.tracker, requested_size, f"malloc({requested_size})")
        return False


class CallocEnterBreakpoint(gdb.Breakpoint):
    def __init__(self, address, tracker) -> None:
        super().__init__(f"*{address:#x}", internal=True)
        self.tracker = tracker

    def stop(self) -> bool:
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)

        num_elements = pwndbg.arguments.argument(0)
        element_size = pwndbg.arguments.argument(1)
        requested_size = element_size * num_elements
        if self.tracker.is_performing_memory_management():
            # This call was made from inside another memory management call.
            # Ignore it.
            return False

        self.tracker.enter_memory_management(CALLOC_NAME)
        AllocExitBreakpoint(self.tracker, requested_size, f"calloc({num_elements}, {element_size})")
        return False


def get_chunk(address, requested_size):
    """
    Reads a chunk from a given address.
    """
    ty = pwndbg.aglib.typeinfo.ppvoid
    size = int(pwndbg.aglib.memory.get_typed_pointer_value(ty, address - ty.sizeof))

    # GLibc bakes the chunk flags in the lowest 3 bits of the size value,
    # so, we separate them here.
    FLAGS_BITMASK = 7

    flags = size & FLAGS_BITMASK
    size ^= flags

    return Chunk(address, size, requested_size, flags)


class AllocExitBreakpoint(gdb.FinishBreakpoint):
    def __init__(self, tracker, requested_size, name) -> None:
        super().__init__(internal=True)
        self.requested_size = requested_size
        self.tracker = tracker
        self.name = name

    def stop(self) -> bool:
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        if not in_program_code_stack():
            # Untracked.
            self.tracker.exit_memory_management()
            return False

        ret_ptr = int(self.return_value)
        if ret_ptr == 0:
            # No change.
            self.tracker.exit_memory_management()
            return False

        chunk = get_chunk(ret_ptr, self.requested_size)
        self.tracker.malloc(chunk)
        ptr_str = self.tracker.colorize_ptr(ret_ptr)
        print(f"[*] {self.name} -> {ptr_str}, {chunk.size:#x} bytes real size")

        self.tracker.exit_memory_management()
        return False

    def out_of_scope(self) -> None:
        print(
            message.warn(
                f"warning: could not follow allocation request of {self.requested_size} bytes"
            )
        )
        self.tracker.exit_memory_management()


class ReallocEnterBreakpoint(gdb.Breakpoint):
    def __init__(self, address, tracker) -> None:
        super().__init__(f"*{address:#x}", internal=True)
        self.tracker = tracker

    def stop(self) -> bool:
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)

        freed_pointer = pwndbg.arguments.argument(0)
        requested_size = pwndbg.arguments.argument(1)
        if self.tracker.is_performing_memory_management():
            # This call was made from inside another memory management call.
            # Ignore it.
            return False

        self.tracker.enter_memory_management(REALLOC_NAME)

        if requested_size == 0:
            # There's no right way to handle realloc(..., 0). C23 says it's
            # undefined behavior, and prior versions say it's implementation-
            # defined. Either way, print a warning and do nothing.
            ptr_str = self.tracker.colorize_ptr(self.freed_pointer)
            print(
                message.warn(
                    f"[-] realloc({ptr_str}, {requested_size}) ignored, as realloc(0, ...) is implementation defined"
                )
            )
            return False

        if freed_pointer == 0:
            # Treat this realloc same as malloc
            AllocExitBreakpoint(self.tracker, requested_size, f"realloc(0x0, {requested_size})")
        else:
            ReallocExitBreakpoint(self.tracker, freed_pointer, requested_size)
        return False


class ReallocExitBreakpoint(gdb.FinishBreakpoint):
    def __init__(self, tracker, freed_ptr, requested_size) -> None:
        super().__init__(internal=True)
        self.freed_ptr = freed_ptr
        self.freed_str = tracker.colorize_ptr(self.freed_ptr)
        self.requested_size = requested_size
        self.tracker = tracker

    def stop(self):
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        if not in_program_code_stack():
            # Untracked.
            self.tracker.exit_memory_management()
            return False

        # Figure out what the reallocated pointer is.
        ret_ptr = int(self.return_value)
        if ret_ptr == 0:
            # No change.
            malloc = None
        chunk = get_chunk(ret_ptr, self.requested_size)
        malloc = lambda: self.tracker.malloc(chunk)

        if not self.tracker.free(self.freed_ptr):
            # This is a chunk we'd never seen before.
            malloc()
            self.tracker.exit_memory_management()

            msg = f"realloc() to {self.requested_size} bytes with previously unknown pointer {self.freed_str}"
            print(f"[!] {msg}")

            global stop_on_error
            if stop_on_error:
                global last_issue
                last_issue = message.error(msg)
            return stop_on_error

        malloc()
        self.tracker.exit_memory_management()

        print(
            f"[*] realloc({self.freed_str}, {self.requested_size}) -> {ret_ptr:#x}, {chunk.size:#x} bytes real size"
        )
        return False

    def out_of_scope(self) -> None:
        print(message.warn(f"warning: could not follow free request for chunk {self.freed_str}"))
        self.tracker.exit_memory_management()


class FreeEnterBreakpoint(gdb.Breakpoint):
    def __init__(self, address, tracker) -> None:
        super().__init__(f"*{address:#x}", internal=True)
        self.tracker = tracker

    def stop(self) -> bool:
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        ptr = pwndbg.arguments.argument(0)
        if self.tracker.is_performing_memory_management():
            # This call was made from inside another memory management call.
            # Ignore it.
            return False
        if ptr == 0:
            # free(0) is a no-op.
            print("[*] free(0x0)")
            return False

        self.tracker.enter_memory_management(FREE_NAME)
        FreeExitBreakpoint(self.tracker, ptr)
        return False


class FreeExitBreakpoint(gdb.FinishBreakpoint):
    def __init__(self, tracker, ptr) -> None:
        super().__init__(internal=True)
        self.ptr = ptr
        self.ptr_str = tracker.colorize_ptr(self.ptr)
        self.tracker = tracker

    def stop(self):
        pwndbg.lib.cache.clear_cache(CacheUntilEvent.STOP)
        if not in_program_code_stack():
            # Untracked.
            self.tracker.exit_memory_management()
            return False

        if not self.tracker.free(self.ptr):
            # This is a chunk we'd never seen before.
            self.tracker.exit_memory_management()

            msg = f"free() with previously unknown pointer {self.ptr_str}"
            print(f"[!] {msg}")
            global stop_on_error
            if stop_on_error:
                global last_issue
                last_issue = message.error(msg)
            return stop_on_error

        self.tracker.exit_memory_management()

        print(f"[*] free({self.ptr_str})")
        return False

    def out_of_scope(self) -> None:
        print(message.warn(f"warning: could not follow free request for chunk {self.ptr_str}"))
        self.tracker.exit_memory_management()


def in_program_code_stack() -> bool:
    exe = pwndbg.aglib.proc.exe()
    binary_exec_page_ranges = tuple(
        (p.start, p.end) for p in pwndbg.aglib.vmmap.get() if p.objfile == exe and p.execute
    )

    frame = gdb.newest_frame()
    while frame is not None:
        pc = frame.pc()
        for start, end in binary_exec_page_ranges:
            if start <= pc < end:
                return True
        frame = frame.older()
    return False


# These variables track the currently installed heap tracker.
malloc_enter = None
calloc_enter = None
realloc_enter = None
free_enter = None

# Whether the inferior should be stopped when an error is detected.
stop_on_error = True


def install(disable_hardware_watchpoints=True) -> None:
    global malloc_enter
    global calloc_enter
    global realloc_enter
    global free_enter

    if is_enabled():
        print("Nothing to do.")
        return

    # Make sure the required functions are available.
    required_symbols = [MALLOC_NAME, FREE_NAME]
    available = [resolve_address(name) for name in required_symbols]

    if not all(available):
        print(message.error("The following required symbols are not available:"))
        for name in (x[0] for x in zip(required_symbols, available) if not x[1]):
            print(message.error(f"    - {name}"))
        print(message.error(f"Make sure {LIBC_NAME} has already been loaded."))

        return

    # Warn our users that this is still an experimental feature and that due to
    # limitations in how GDB handles breakpoint creation and deletion during
    # processing of stop events for other breakpoints, there's not a lot we can
    # do about it currently.
    #
    # See https://sourceware.org/pipermail/gdb/2024-January/051062.html
    print(
        message.warn(
            "This feature is experimental and is known to report false positives, take the"
        )
    )
    print(message.warn("diagnostics it produces with a grain of salt. Use at your own risk."))
    print()

    # Disable hardware watchpoints.
    #
    # We don't really know how to make sure that the hardware watchpoints
    # present in the system have enough capabilities for them to be useful to
    # us in this module, seeing as what they can do varies considerably between
    # systems and failures are fairly quiet and, thus, hard to detect[1].
    # Because of this, we opt to disable them by default for the sake of
    # consistency and so that we don't have to chase silent failures.
    #
    # [1]: https://sourceware.org/gdb/onlinedocs/gdb/Set-Watchpoints.html
    if disable_hardware_watchpoints:
        gdb.execute("set can-use-hw-watchpoints 0")
        print("Hardware watchpoints have been disabled. Please do not turn them back on until")
        print("heap tracking is disabled, as it may lead to unexpected silent errors.")
        print()
        print("They may be re-enabled with `set can-use-hw-watchpoints 1`")
        print()
    else:
        print(
            message.warn("Hardware watchpoints have not been disabled, silent errors may happen.")
        )
        print()

    # Install the heap tracker.
    tracker = Tracker()

    malloc_enter = MallocEnterBreakpoint(available[0], tracker)
    free_enter = FreeEnterBreakpoint(available[1], tracker)

    calloc_address = resolve_address(CALLOC_NAME)
    if calloc_address:
        calloc_enter = CallocEnterBreakpoint(calloc_address, tracker)

    realloc_address = resolve_address(REALLOC_NAME)
    if realloc_address:
        realloc_enter = ReallocEnterBreakpoint(realloc_address, tracker)

    print("Heap tracker installed.")


def uninstall() -> None:
    global malloc_enter
    global calloc_enter
    global realloc_enter
    global free_enter

    if is_enabled():
        malloc_enter.delete()
        free_enter.delete()

        malloc_enter = None
        free_enter = None

        if calloc_enter is not None:
            calloc_enter.delete()
            calloc_enter = None
        if realloc_enter is not None:
            realloc_enter.delete()
            realloc_enter = None

        print("Heap tracker removed.")
    else:
        print("Nothing to do.")



from __future__ import annotations

import fnmatch
import os.path
import re
from typing import Any

import pygments
import pygments.formatters
import pygments.lexers
import pygments.util
from pwnlib.lexer import PwntoolsLexer

import pwndbg
import pwndbg.lib.cache
from pwndbg.color import disable_colors
from pwndbg.color import message
from pwndbg.color import theme

pwndbg.config.add_param("syntax-highlight", True, "source code / assembly syntax highlight")
style = theme.add_param(
    "syntax-highlight-style",
    "monokai",
    "source code / assembly syntax highlight stylename of pygments module",
)

formatter = pygments.formatters.Terminal256Formatter(style=str(style))
pwntools_lexer = PwntoolsLexer()
lexer_cache: dict[str, Any] = {}


@pwndbg.config.trigger(style)
def check_style() -> None:
    global formatter
    try:
        formatter = pygments.formatters.Terminal256Formatter(style=str(style))

        # Reset the highlighted source cache
        from pwndbg.commands.context import get_highlight_source

        pwndbg.lib.cache.clear_function_cache(get_highlight_source)
    except pygments.util.ClassNotFound:
        print(
            message.warn(f"The pygment formatter style '{style}' is not found, restore to default")
        )
        style.revert_default()


def _fn_matches(filename, pattern):
    # faster alternative to the naive regex matching algorithm that pygments uses
    # most of the regexs are of the form "*.<extension>", which can be converted to a simple string match
    extension = pattern[2:]
    if pattern.isascii() and pattern[0] == "*" and pattern[1] == "." and extension.isalnum():
        # to avoid an extra string copy, we also need to check whether the filename has a '.' before the extension
        return filename.endswith(extension) and filename[-len(extension) - 1] == "."

    # fall back to slow regex
    return re.match(fnmatch.translate(pattern), filename)


def _pygments_get_lexer_for_filename(filename, code, **options):
    """
    A faster alternative to pygment's get_lexer_for_filename only checks the
    built-in lexers unless a match is not found, in which case it falls back to
    pygment's get_lexer_for_filename, which also checks for plugin lexers.
    """

    # fall back to slow method if there are multiple matches
    one_match = False
    matched_lexer = ""
    fn = os.path.basename(filename)
    for name, _, filenames, _ in pygments.lexers.get_all_lexers(plugins=False):
        for candidate_filename in filenames:
            if _fn_matches(fn, candidate_filename):
                if one_match:
                    # already seen one match, this is a second match
                    one_match = False
                    break
                one_match = True
                matched_lexer = name
    if one_match:
        return pygments.lexers.get_lexer_by_name(matched_lexer, **options)
    # either we can't find it or there are multiple matches to choose from
    return pygments.lexers.guess_lexer_for_filename(filename, code, **options)


def syntax_highlight(code: str, filename: str = ".asm") -> str:
    # No syntax highlight if pygment is not installed
    if disable_colors:
        return code

    filename = os.path.basename(filename)

    lexer = lexer_cache.get(filename, None)

    # If source code is asm, use our customized lexer.
    # Note: We can not register our Lexer to pygments and use their APIs,
    # since the pygment only search the lexers installed via setuptools.
    if not lexer:
        for glob_pat in PwntoolsLexer.filenames:
            pat = "^" + glob_pat.replace(".", r"\.").replace("*", r".*") + "$"
            if re.match(pat, filename):
                lexer = pwntools_lexer
                break

    if not lexer:
        try:
            lexer = _pygments_get_lexer_for_filename(filename, code, stripnl=False)
        except pygments.util.ClassNotFound:
            # no lexer for this file or invalid style
            pass

    if lexer:
        lexer_cache[filename] = lexer

        code = pygments.highlight(code, lexer, formatter).rstrip()

    return code



"""
Stepping until an event occurs
"""

from __future__ import annotations

import argparse

import pwndbg.aglib.next
import pwndbg.aglib.proc
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.dbg_mod
from pwndbg.commands import CommandCategory


async def _nextjmp(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `nextjmp` command.
    """

    if await pwndbg.aglib.next.break_next_branch(ec):
        pwndbg.commands.context.context()


@pwndbg.commands.Command(
    "Breaks at the next jump instruction.", aliases=["nextjump"], category=CommandCategory.NEXT
)
@pwndbg.commands.OnlyWhenRunning
def nextjmp() -> None:
    """Breaks at the next jump instruction"""

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_nextjmp)


parser = argparse.ArgumentParser(description="Breaks at the next call instruction.")
parser.add_argument(
    "symbol_regex",
    type=str,
    default=None,
    nargs="?",
    help="A regex matching the name of next symbol to be broken on before calling.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.NEXT)
@pwndbg.commands.OnlyWhenRunning
def nextcall(symbol_regex=None) -> None:
    """Breaks at the next call instruction"""

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        if await pwndbg.aglib.next.break_next_call(ec, symbol_regex):
            pwndbg.commands.context.context()

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)


async def _nextret(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `nextret` command.
    """
    if await pwndbg.aglib.next.break_next_ret(ec):
        pwndbg.commands.context.context()


@pwndbg.commands.Command("Breaks at next return-like instruction.", category=CommandCategory.NEXT)
@pwndbg.commands.OnlyWhenRunning
def nextret() -> None:
    """Breaks at next return-like instruction"""

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_nextret)


async def _stepret(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `stepret` command.
    """
    while (
        pwndbg.aglib.proc.alive()
        and not (await pwndbg.aglib.next.break_next_ret(ec))
        and (await pwndbg.aglib.next.break_next_branch(ec))
    ):
        # Here we are e.g. on a CALL instruction (temporarily breakpointed by `break_next_branch`)
        # We need to step so that we take this branch instead of ignoring it
        await ec.single_step()
        continue


@pwndbg.commands.Command(
    "Breaks at next return-like instruction by 'stepping' to it.", category=CommandCategory.NEXT
)
@pwndbg.commands.OnlyWhenRunning
def stepret() -> None:
    """Breaks at next return-like instruction by 'stepping' to it"""

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_stepret)


async def _nextproginstr(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `nextproginstr` command.
    """
    await pwndbg.aglib.next.break_on_program_code(ec)


@pwndbg.commands.Command(
    "Breaks at the next instruction that belongs to the running program.",
    category=CommandCategory.NEXT,
)
@pwndbg.commands.OnlyWhenRunning
def nextproginstr() -> None:
    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_nextproginstr)


parser = argparse.ArgumentParser(description="Breaks on the instruction after this one.")
parser.add_argument("addr", type=int, default=None, nargs="?", help="The address to break after.")


@pwndbg.commands.Command(parser, aliases=["so"], category=CommandCategory.NEXT)
@pwndbg.commands.OnlyWhenRunning
def stepover(addr=None) -> None:
    """Sets a breakpoint on the instruction after this one"""

    async def _stepover(ec: pwndbg.dbg_mod.ExecutionController):
        """
        Execution controller for the `stepover` command.
        """
        await pwndbg.aglib.next.break_on_next(ec, addr)

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_stepover)


async def _nextsyscall(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `nextsyscall` command
    """
    while (
        pwndbg.aglib.proc.alive()
        and not (await pwndbg.aglib.next.break_next_interrupt(ec))
        and (await pwndbg.aglib.next.break_next_branch(ec))
    ):
        continue


@pwndbg.commands.Command(
    "Breaks at the next syscall not taking branches.",
    aliases=["nextsc"],
    category=CommandCategory.NEXT,
)
@pwndbg.commands.OnlyWhenRunning
def nextsyscall() -> None:
    """
    Breaks at the next syscall not taking branches.
    """
    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_nextsyscall)


async def _stepsyscall(ec: pwndbg.dbg_mod.ExecutionController):
    """
    Execution controller for the `stepsyscall` command.
    """

    while (
        pwndbg.aglib.proc.alive()
        and not (await pwndbg.aglib.next.break_next_interrupt(ec, honor_current_branch=True))
        and (await pwndbg.aglib.next.break_next_branch(ec, including_current=True))
    ):
        # Here we are e.g. on a CALL instruction (temporarily breakpointed by `break_next_branch`)
        # We need to step so that we take this branch instead of ignoring it
        await ec.single_step()
        continue


@pwndbg.commands.Command(
    "Breaks at the next syscall by taking branches.",
    aliases=["stepsc"],
    category=CommandCategory.NEXT,
)
@pwndbg.commands.OnlyWhenRunning
def stepsyscall() -> None:
    """
    Breaks at the next syscall by taking branches.
    """

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(_stepsyscall)


parser = argparse.ArgumentParser(description="Breaks on the next matching instruction.")
parser.add_argument("mnemonic", type=str, help="The mnemonic of the instruction")
parser.add_argument(
    "op_str",
    type=str,
    nargs="*",
    help="The operands of the instruction",
)


@pwndbg.commands.Command(parser, category=CommandCategory.NEXT)
@pwndbg.commands.OnlyWhenRunning
def stepuntilasm(mnemonic, op_str) -> None:
    if len(op_str) == 0:
        op_str = None

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        await pwndbg.aglib.next.break_on_next_matching_instruction(ec, mnemonic, op_str)

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)



#!/usr/bin/env python3
from __future__ import annotations

import os
import shutil
import site
import subprocess
import sys
import sysconfig


def get_gdb_version(path: str) -> tuple[str, ...]:
    result = subprocess.run(
        [
            path,
            "-nx",
            "--batch",
            "-iex",
            "py import sysconfig; print(sysconfig.get_config_var('INSTSONAME'), sysconfig.get_config_var('VERSION'))",
        ],
        capture_output=True,
        text=True,
    )

    arr = result.stdout.strip().split(" ", 1)
    if len(arr) != 2:
        return "", ""
    return arr[0], arr[1]


def get_venv_bin_path() -> str:
    bin_dir = "Scripts" if os.name == "nt" else "bin"
    return os.path.join(sys.prefix, bin_dir)


def prepend_venv_bin_to_path():
    # Set virtualenv's bin path (needed for utility tools like ropper, pwntools etc)
    venv_bin = get_venv_bin_path()
    path_elements = os.environ.get("PATH", "").split(os.pathsep)
    if venv_bin in path_elements:
        return

    path_elements.insert(0, venv_bin)
    os.environ["PATH"] = os.pathsep.join(path_elements)


def main():
    prepend_venv_bin_to_path()

    gdb_argv = [
        sys.argv[0],
        "-q",
        "-nx",
        "-iex",
        "py import pwndbginit.gdbinit; pwndbginit.gdbinit.main_try()",
        *sys.argv[1:],
    ]
    sys.argv = gdb_argv

    try:
        from gdb_for_pwndbg.gdb import main  # type: ignore[import-untyped]

        main()
        return
    except ImportError:
        pass

    gdb_path = shutil.which("gdb")
    if not gdb_path:
        print("ERROR: Could not find 'gdb' binary")
        sys.exit(1)

    envs = os.environ.copy()
    envs["PYTHONNOUSERSITE"] = "1"
    envs["PYTHONPATH"] = ":".join(site.getsitepackages())

    # Ensure arg0 points to the gdb binary; otherwise GDB can pick up a wrong PYTHONHOME.
    sys.argv[0] = gdb_path

    # sys.prefix/sys.exec_prefix must point to the virtual environment,
    # otherwise our auto-upgrade mechanism won't work when the package is installed in editable mode
    prefix_cmd = (
        f"py import sys; sys.prefix = {sys.prefix!r}; sys.exec_prefix = {sys.exec_prefix!r}"
    )
    sys.argv.insert(1, prefix_cmd)
    sys.argv.insert(1, "-iex")

    expected = (sysconfig.get_config_var("INSTSONAME"), sysconfig.get_config_var("VERSION"))
    have = get_gdb_version(gdb_path)
    if have != expected:
        print(
            f"ERROR: GDB is compiled for Python {have}, but your Python interpreter is {expected} .\n\n"
            "If you installed Pwndbg with `uv tool install` and want to use system GDB you may try:\n"
            """  PY_VER=$(gdb -nx --batch -iex 'py import sysconfig; print(sysconfig.get_config_var("VERSION"))')\n"""
            "  uv tool install --python=$PY_VER .\n\n"
            "(add --editable if you are doing development)\n"
            "(the python version may be wrong because our development scripts switch between them)"
        )
        sys.exit(1)

    os.execve(gdb_path, sys.argv, env=envs)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.mips import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib.disasm.arch
import pwndbg.color.memory as mem_color
import pwndbg.dintegration
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.arch import register_assign
from pwndbg.aglib.disasm.instruction import FORWARD_JUMP_GROUP
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

# These branch instructions do not have a branch delay
BRANCH_WITHOUT_DELAY_SLOT_INSTRUCTIONS = {
    MIPS_INS_BC,
    MIPS_INS_BALC,
    MIPS_INS_JIALC,
    MIPS_INS_JIC,
    MIPS_INS_BLEZALC,
    MIPS_INS_BGEZALC,
    MIPS_INS_BGTZALC,
    MIPS_INS_BLTZALC,
    MIPS_INS_BEQZALC,
    MIPS_INS_BNEZALC,
    MIPS_INS_BLEZC,
    MIPS_INS_BGEZC,
    MIPS_INS_BGEUC,
    MIPS_INS_BGEIC,
    MIPS_INS_BGEUC,
    MIPS_INS_BGEIUC,
    MIPS_INS_BGTZC,
    MIPS_INS_BLTZC,
    MIPS_INS_BEQZC,
    MIPS_INS_ALIAS_BEQZC,
    MIPS_INS_BNEZC,
    MIPS_INS_ALIAS_BNEZC,
    MIPS_INS_BEQC,
    MIPS_INS_ALIAS_BEQC,
    MIPS_INS_BEQIC,
    MIPS_INS_BNEC,
    MIPS_INS_ALIAS_BNEC,
    MIPS_INS_BNEIC,
    MIPS_INS_BLTC,
    MIPS_INS_BLTIC,
    MIPS_INS_BLTUC,
    MIPS_INS_BLTIUC,
    MIPS_INS_BGEC,
    MIPS_INS_BLTUC,
    MIPS_INS_BNVC,
    MIPS_INS_BOVC,
    MIPS_INS_BRSC,
    MIPS_INS_BALRSC,
    MIPS_INS_BBEQZC,
    MIPS_INS_BBNEZC,
}

# Instruction in branch delay slot is ignored if the branch it NOT taken
# Currently unused
BRANCH_LIKELY_INSTRUCTIONS = {
    MIPS_INS_BC1FL,
    MIPS_INS_ALIAS_BC1FL,
    MIPS_INS_BC1TL,
    MIPS_INS_ALIAS_BC1TL,
    MIPS_INS_BEQL,
    MIPS_INS_BGEZALL,
    MIPS_INS_BGEZL,
    MIPS_INS_BGTZL,
    MIPS_INS_BLEZL,
    MIPS_INS_BLTZALL,
    MIPS_INS_BLTZL,
    MIPS_INS_BNEL,
    MIPS_INS_ALIAS_BNEZL,
    MIPS_INS_ALIAS_BEQZL,
}

CONDITION_RESOLVERS: dict[int, Callable[[list[int]], bool]] = {
    MIPS_INS_BEQZ: lambda ops: ops[0] == 0,
    MIPS_INS_BNEZ: lambda ops: ops[0] != 0,
    MIPS_INS_BEQ: lambda ops: ops[0] == ops[1],
    MIPS_INS_BNE: lambda ops: ops[0] != ops[1],
    MIPS_INS_BGEZ: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) >= 0,
    MIPS_INS_BGEZAL: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) >= 0,
    MIPS_INS_BGTZ: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) > 0,
    MIPS_INS_BLEZ: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) <= 0,
    MIPS_INS_BLTZAL: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) < 0,
    MIPS_INS_BLTZ: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits) < 0,
}

CONDITION_RESOLVERS[MIPS_INS_ALIAS_BEQZ] = CONDITION_RESOLVERS[MIPS_INS_BEQZ]
CONDITION_RESOLVERS[MIPS_INS_ALIAS_BNEZ] = CONDITION_RESOLVERS[MIPS_INS_BNEZ]

# These are instructions that have the first operand as the destination register.
# They all do some computation and set the register to the result.
# These were derived from "MIPS Architecture for Programmers Volume II: The MIPS64 Instruction Set Reference Manual"
MIPS_SIMPLE_DESTINATION_INSTRUCTIONS = {
    MIPS_INS_CLO,
    MIPS_INS_CLZ,
    MIPS_INS_DCLO,
    MIPS_INS_DCLZ,
    MIPS_INS_LSA,
    MIPS_INS_DLSA,
    MIPS_INS_MFHI,
    MIPS_INS_MFLO,
    MIPS_INS_SEB,
    MIPS_INS_SEH,
    MIPS_INS_WSBH,
    MIPS_INS_SLT,
    MIPS_INS_SLTI,
    MIPS_INS_SLTIU,
    MIPS_INS_SLTU,
    MIPS_INS_MOVN,
    # Rare - unaligned read - have complex loading logic
    MIPS_INS_LDL,
    MIPS_INS_LDR,
    # Rare - partial load on portions of address
    MIPS_INS_LWL,
    MIPS_INS_LWR,
}

# All MIPS load instructions
MIPS_LOAD_INSTRUCTIONS = {
    MIPS_INS_LB: -1,
    MIPS_INS_LBU: 1,
    MIPS_INS_LH: -2,
    MIPS_INS_LHU: 2,
    MIPS_INS_LW: -4,
    MIPS_INS_LWU: 4,
    MIPS_INS_LWPC: -4,
    MIPS_INS_LWUPC: 4,
    MIPS_INS_LD: -8,
    MIPS_INS_LDPC: 8,
}

MIPS_STORE_INSTRUCTIONS = {
    MIPS_INS_SB: 1,
    MIPS_INS_SH: 2,
    MIPS_INS_SW: 4,
    MIPS_INS_SD: 8,
}

MIPS_BINARY_OPERATIONS = {
    MIPS_INS_ADD: "+",
    MIPS_INS_ADDI: "+",
    MIPS_INS_ADDIU: "+",
    MIPS_INS_ADDU: "+",
    MIPS_INS_DADD: "+",
    MIPS_INS_DADDI: "+",
    MIPS_INS_DADDIU: "+",
    MIPS_INS_DADDU: "+",
    MIPS_INS_SUB: "-",
    MIPS_INS_SUBU: "-",
    MIPS_INS_DSUB: "-",
    MIPS_INS_DSUBU: "-",
    MIPS_INS_ANDI: "&",
    MIPS_INS_AND: "&",
    MIPS_INS_ORI: "|",
    MIPS_INS_OR: "|",
    MIPS_INS_XOR: "^",
    MIPS_INS_XORI: "^",
    MIPS_INS_SLL: "<<",
    MIPS_INS_SLLV: "<<",
    MIPS_INS_DSLL: "<<",
    MIPS_INS_DSLLV: "<<",
    MIPS_INS_SRL: ">>",
    MIPS_INS_SRLV: ">>",
    MIPS_INS_DSRL: ">>",
    MIPS_INS_DSRLV: ">>",
}


# This class enhances 32-bit, 64-bit, and micro MIPS
class MipsDisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    def __init__(self, architecture) -> None:
        super().__init__(architecture)

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {
            # MOVE
            MIPS_INS_MOVE: self._common_move_annotator,
            MIPS_INS_ALIAS_MOVE: self._common_move_annotator,
            # LI
            MIPS_INS_LI: self._common_move_annotator,
            # LUI
            MIPS_INS_LUI: self._lui_annotator,
        }

    @override
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        if instruction.id in MIPS_LOAD_INSTRUCTIONS:
            read_size = MIPS_LOAD_INSTRUCTIONS[instruction.id]

            self._common_load_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                abs(read_size),
                read_size < 0,
                pwndbg.aglib.arch.ptrsize,
                instruction.operands[0].str,
                instruction.operands[1].str,
            )
        elif instruction.id in MIPS_STORE_INSTRUCTIONS:
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                instruction.operands[0].before_value,
                MIPS_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[1].str,
            )
        elif instruction.id in MIPS_BINARY_OPERATIONS:
            self._common_binary_op_annotator(
                instruction,
                emu,
                instruction.operands[0],
                instruction.operands[1].before_value,
                instruction.operands[2].before_value,
                MIPS_BINARY_OPERATIONS[instruction.id],
            )
        elif instruction.id in MIPS_SIMPLE_DESTINATION_INSTRUCTIONS:
            self._common_generic_register_destination(instruction, emu)
        else:
            self.annotation_handlers.get(instruction.id, lambda *a: None)(instruction, emu)

    def _lui_annotator(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        result_operand, right = instruction.operands
        if result_operand.str and right.before_value is not None:
            if (address := result_operand.after_value) is None:
                # Resolve it manually without emulation
                address = right.before_value << 16

            instruction.annotation = register_assign(
                result_operand.str,
                mem_color.get_address_and_symbol(
                    address, pwndbg.dintegration.manager.get_stack_var_dict_all()
                ),
            )

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        if len(instruction.operands) == 0:
            return InstructionCondition.UNDETERMINED

        # Not using list comprehension because they run in a separate scope in which super() does not exist
        resolved_operands: list[int] = []
        for op in instruction.operands:
            resolved_operands.append(
                super()._resolve_used_value(op.before_value, instruction, op, emu)
            )

        # If any of the relevent operands are None (we can't reason about them), quit.
        if any(value is None for value in resolved_operands[:-1]):
            # Note the [:-1]. MIPS jump instructions have the target as the last operand
            # https://www.doc.ic.ac.uk/lab/secondyear/spim/node16.html
            return InstructionCondition.UNDETERMINED

        conditional = CONDITION_RESOLVERS.get(instruction.id, lambda *a: None)(resolved_operands)

        if conditional is None:
            return InstructionCondition.UNDETERMINED

        return InstructionCondition.TRUE if conditional else InstructionCondition.FALSE

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        if bool(instruction.groups & FORWARD_JUMP_GROUP) and not bool(
            instruction.groups & BRANCH_WITHOUT_DELAY_SLOT_INSTRUCTIONS
        ):
            instruction.causes_branch_delay = True

        return super()._resolve_target(instruction, emu)

    @override
    def _parse_memory(
        self,
        instruction: PwndbgInstruction,
        op: pwndbg.aglib.disasm.arch.EnhancedOperand,
        emu: Emulator,
    ) -> int | None:
        """
        Parse the `MipsOpMem` Capstone object to determine the concrete memory address used.
        """
        base = self._read_register(instruction, op.mem.base, emu)
        if base is None:
            return None
        return base + op.mem.disp



"""
Search the address space for byte patterns.
"""

from __future__ import annotations

from collections.abc import Collection
from collections.abc import Generator

import pwndbg.aglib.vmmap
import pwndbg.lib.memory


def search(
    searchfor: bytes | bytearray,
    mappings: Collection[pwndbg.lib.memory.Page] | None = None,
    start: int | None = None,
    end: int | None = None,
    step: int | None = None,
    aligned: int | None = None,
    limit: int | None = None,
    executable: bool = False,
    writable: bool = False,
) -> Generator[int, None, None]:
    """Search inferior memory for a byte sequence.

    Arguments:
        searchfor: Byte sequence to find
        mappings: List of pwndbg.lib.memory.Page objects to search
            By default, uses all available mappings.
        start: First address to search, inclusive.
        end: Last address to search, exclusive.
        step: Size of memory region to skip each result
        aligned: Strict byte alignment for search result
        limit: Maximum number of results to return
        executable: Restrict search to executable pages
        writable: Restrict search to writable pages

    Yields:
        An iterator on the address matches
    """
    inf = pwndbg.dbg.selected_inferior()

    maps = mappings or pwndbg.aglib.vmmap.get()

    if end and start:
        assert start < end, "Last address to search must be greater then first address"
        maps = [m for m in maps if start in m or (end - 1) in m]
    elif start:
        maps = [m for m in maps if start in m]
    elif end:
        maps = [m for m in maps if (end - 1) in m]

    if executable:
        maps = [m for m in maps if m.execute]

    if writable:
        maps = [m for m in maps if m.write]

    if len(maps) == 0:
        print("No applicable memory regions found to search in.")
        return

    count = 0
    if limit and limit <= 0:
        return

    for vmmap in maps:
        start = vmmap.start
        end = vmmap.end

        if limit and count >= limit:
            break

        for element in inf.find_in_memory(
            bytearray(searchfor),
            start,
            end - start,
            aligned or 1,
            (limit - count) if limit else -1,
            step or -1,
        ):
            yield element
            count += 1



from __future__ import annotations

import argparse
import subprocess

import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.rizin
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Launches rizin.")
parser.add_argument("--no-seek", action="store_true", help="Do not seek to current pc")
parser.add_argument(
    "--no-rebase",
    action="store_true",
    help="Do not set the base address for PIE according to the current mapping",
)
parser.add_argument("arguments", nargs="*", type=str, help="Arguments to pass to rizin")


@pwndbg.commands.Command(
    parser,
    aliases=["rizin"],
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> rz -- -AA
[x] Analyze all flags starting with sym. and entry0 (aa)
[x] Analyze function calls
[...]
[x] Enable constraint types analysis for variables
 -- Use 'e asm.offset=true' to show offsets in 16bit segment addressing mode.
[0x0001d3d0]>
    """,
)
@pwndbg.commands.OnlyWithFile
def rz(arguments: list[str], no_seek: bool = False, no_rebase: bool = False) -> None:
    filename = pwndbg.aglib.file.get_proc_exe_file()

    # Build up the command line to run
    cmd = ["rizin"]
    flags = ["-e", "io.cache=true"]
    exe = pwndbg.aglib.elf.exe()
    if pwndbg.aglib.proc.alive() and exe is not None:
        addr = pwndbg.aglib.regs.pc
        if pwndbg.aglib.elf.get_elf_info(filename).is_pie:
            if no_rebase:
                addr -= exe.address
            else:
                flags.extend(["-B", hex(exe.address)])
        if not no_seek:
            cmd.extend(["-s", hex(addr)])
    cmd.extend(flags)
    cmd += arguments
    cmd.extend([filename])

    try:
        subprocess.call(cmd)
    except Exception:
        print("Could not run rizin. Please ensure it's installed and in $PATH.")


parser = argparse.ArgumentParser(
    description="Execute stateful rizin commands through rzpipe.",
)
parser.add_argument("arguments", nargs="+", type=str, help="Arguments to pass to rzpipe")


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> rzpipe pdf @ sym.main
            ; DATA XREF from entry0 @ 0x1d3e8
┌ int main(int argc, char **argv, char **envp);
│           ; arg int argc @ rdi
│           ; arg char **argv @ rsi
│           ; arg char **envp @ rdx
[...]
│           ; var uint64_t var_18ch @ stack - 0x18c
│           ; var int64_t var_188h @ stack - 0x188
[...]
│           ; var int64_t canary @ stack - 0x40
│           0x0001b920      endbr64
│           0x0001b924      push  r15
│           0x0001b926      push  r14
│           0x0001b928      push  r13
│           0x0001b92a      push  r12
[...]
│           0x0001b966      test  eax, eax
│       ┌─< 0x0001b968      jne   0x1b9af
│       │   0x0001b96a      call  sym.xtrace_init
│       │   0x0001b96f      call  sym.check_dev_tty
│       │   ; CODE XREF from main @ 0x1b988
│      ┌──> 0x0001b974      cmp   dword [obj.debugging_login_shell], 0 ; [0x11224c:4]=0
│     ┌───< 0x0001b97b      je    0x1b9ba
│     │╎│   0x0001b97d      mov   edi, 3                               ; int s
[...]
    """,
)
@pwndbg.commands.OnlyWithFile
def rzpipe(arguments: list[str]) -> None:
    print(pwndbg.aglib.rizin.rzcmd(arguments))



"""
This file declares types and methods useful for enumerating
all of the address spaces and permissions of an ELF file in memory.

This is necessary for when access to /proc is restricted, or when
working on a BSD system which simply does not have /proc.
"""

from __future__ import annotations

import ctypes
import importlib
import subprocess
import sys
from typing import NamedTuple
from typing import TypeAlias
from typing import TypeVar

from elftools.elf.constants import SH_FLAGS
from elftools.elf.elffile import ELFFile
from elftools.elf.relocation import Relocation
from elftools.elf.relocation import RelocationSection

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.file
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.auxv
import pwndbg.dbg_mod
import pwndbg.lib
import pwndbg.lib.cache
import pwndbg.lib.config
import pwndbg.lib.elftypes
import pwndbg.lib.memory
import pwndbg.lib.zig
from pwndbg.color import message
from pwndbg.dbg_mod import EventType
from pwndbg.lib import Status

# ELF constants
PF_X, PF_W, PF_R = 1, 2, 4
ET_EXEC, ET_DYN = 2, 3


module = sys.modules[__name__]


class ELFInfo(NamedTuple):
    """
    ELF metadata and structures.
    """

    header: dict[str, int | str]
    sections: list[dict[str, int | str]]
    segments: list[dict[str, int | str]]

    @property
    def is_pic(self) -> bool:
        return self.header["e_type"] == "ET_DYN"

    @property
    def is_pie(self) -> bool:
        return self.is_pic


Ehdr: TypeAlias = pwndbg.lib.elftypes.Elf32_Ehdr | pwndbg.lib.elftypes.Elf64_Ehdr
Phdr: TypeAlias = pwndbg.lib.elftypes.Elf32_Phdr | pwndbg.lib.elftypes.Elf64_Phdr


@pwndbg.dbg.event_handler(EventType.START)
@pwndbg.dbg.event_handler(EventType.NEW_MODULE)
def update() -> None:
    global Ehdr, Phdr
    try:
        importlib.reload(pwndbg.lib.elftypes)

    except ImportError:
        print(message.warn("Failed to reload pwndbg.lib.elftypes"))
        pass

    if pwndbg.aglib.arch.ptrsize == 4:
        Ehdr = pwndbg.lib.elftypes.Elf32_Ehdr
        Phdr = pwndbg.lib.elftypes.Elf32_Phdr
    else:
        Ehdr = pwndbg.lib.elftypes.Elf64_Ehdr
        Phdr = pwndbg.lib.elftypes.Elf64_Phdr

    module.__dict__.update(locals())


T = TypeVar(
    "T",
    pwndbg.lib.elftypes.Elf32_Ehdr | pwndbg.lib.elftypes.Elf64_Ehdr,
    pwndbg.lib.elftypes.Elf32_Phdr | pwndbg.lib.elftypes.Elf64_Phdr,
)


def read(typ: T, address: int, blob: bytearray | None = None) -> T:
    size = ctypes.sizeof(typ)

    if not blob:
        data = pwndbg.aglib.memory.read(address, size)
    else:
        data = blob[address : address + size]

    obj = typ.from_buffer_copy(data)
    obj.address = address
    obj.type = typ
    return obj


@pwndbg.lib.cache.cache_until("objfile")
def get_elf_info(filepath: str) -> ELFInfo:
    """
    Parse and return ELFInfo.

    Adds various calculated properties to the ELF header, segments and sections.
    Such added properties are those with prefix 'x_' in the returned dicts.
    """
    local_path = pwndbg.aglib.file.get_file(filepath)
    with open(local_path, "rb") as f:
        elffile = ELFFile(f)
        header = dict(elffile.header)
        segments = []
        for seg in elffile.iter_segments():
            s = dict(seg.header)
            s["x_perms"] = [
                mnemonic
                for mask, mnemonic in [(PF_R, "read"), (PF_W, "write"), (PF_X, "execute")]
                if s["p_flags"] & mask != 0
            ]
            # end of memory backing
            s["x_vaddr_mem_end"] = s["p_vaddr"] + s["p_memsz"]
            # end of file backing
            s["x_vaddr_file_end"] = s["p_vaddr"] + s["p_filesz"]
            segments.append(s)
        sections = []
        for sec in elffile.iter_sections():
            s = dict(sec.header)
            s["x_name"] = sec.name
            s["x_addr_mem_end"] = s["x_addr_file_end"] = s["sh_addr"] + s["sh_size"]
            sections.append(s)
        return ELFInfo(header, sections, segments)


@pwndbg.lib.cache.cache_until("objfile")
def get_elf_info_rebased(filepath: str, vaddr: int) -> ELFInfo:
    """
    Parse and return ELFInfo with all virtual addresses rebased to vaddr
    """
    raw_info = get_elf_info(filepath)
    # silently ignores "wrong" vaddr supplied for non-PIE ELF
    load = vaddr if raw_info.is_pic else 0
    headers = dict(raw_info.header)
    headers["e_entry"] += load  # type: ignore[operator]

    segments: list[dict[str, int | str]] = []
    for seg in raw_info.segments:
        s = dict(seg)
        for vaddr_attr in ["p_vaddr", "x_vaddr_mem_end", "x_vaddr_file_end"]:
            s[vaddr_attr] += load  # type: ignore[operator]
        segments.append(s)

    sections: list[dict[str, int | str]] = []
    for sec in raw_info.sections:
        s = dict(sec)
        for vaddr_attr in ["sh_addr", "x_addr_mem_end", "x_addr_file_end"]:
            s[vaddr_attr] += load  # type: ignore[operator]
        sections.append(s)

    return ELFInfo(headers, sections, segments)


def get_containing_segments(elf_filepath: str, elf_loadaddr: int, vaddr: int):
    elf = get_elf_info_rebased(elf_filepath, elf_loadaddr)
    segments = []
    for seg in elf.segments:
        # disregard segments which were unable to be named by pyelftools (see #777)
        # and non-LOAD segments that are not file-backed (typically STACK)
        if isinstance(seg["p_type"], int) or ("LOAD" not in seg["p_type"] and seg["p_filesz"] == 0):
            continue
        # disregard segments not containing vaddr
        if vaddr < seg["p_vaddr"] or vaddr >= seg["x_vaddr_mem_end"]:  # type: ignore[operator]
            continue
        segments.append(dict(seg))
    return segments


def get_containing_sections(elf_filepath: str, elf_loadaddr: int, vaddr: int):
    elf = get_elf_info_rebased(elf_filepath, elf_loadaddr)
    sections = []
    for sec in elf.sections:
        # disregard sections not occupying memory
        if sec["sh_flags"] & SH_FLAGS.SHF_ALLOC == 0:  # type: ignore[operator]
            continue
        # disregard sections that do not contain vaddr
        if vaddr < sec["sh_addr"] or vaddr >= sec["x_addr_mem_end"]:  # type: ignore[operator]
            continue
        sections.append(dict(sec))
    return sections


def get_vmlinux_unrand_base(elf_filepath: str) -> int | None:
    elf = get_elf_info(elf_filepath)
    for seg in elf.segments:
        if seg["p_type"] == "PT_LOAD":
            return int(seg["p_vaddr"])
    return None


def section_by_name(
    filepath: str, section_name: str, try_local_path: bool = False
) -> tuple[int, int, bytes] | None:
    """
    Return the content of a section from an ELF file, as a (start address, size, content) tuple.
    """
    # TODO: We should have some cache mechanism or something at `pndbg.aglib.file.get_file()` in the future to avoid downloading the same file multiple times when we are debugging a remote process
    local_path = pwndbg.aglib.file.get_file(filepath, try_local_path=try_local_path)

    with open(local_path, "rb") as f:
        elffile = ELFFile(f)
        section = elffile.get_section_by_name(section_name)
        return (section["sh_addr"], section["sh_size"], section.data()) if section else None


def relocations_by_section_name(
    filepath: str, section_name: str, try_local_path: bool = False
) -> tuple[Relocation, ...]:
    """
    Return the relocation entries of a section from an ELF file as a (possibly empty)
    tuple of Relocation objects.
    """
    # TODO: We should have some cache mechanism or something at `pndbg.aglib.file.get_file()` in the future to avoid downloading the same file multiple times when we are debugging a remote process
    local_path = pwndbg.aglib.file.get_file(filepath, try_local_path=try_local_path)

    with open(local_path, "rb") as f:
        elffile = ELFFile(f)
        section = elffile.get_section_by_name(section_name)
        if section is None or not isinstance(section, RelocationSection):
            return ()
        return tuple(section.iter_relocations())


@pwndbg.aglib.proc.OnlyWhenRunning
@pwndbg.lib.cache.cache_until("start", "objfile")
def exe() -> Ehdr | None:
    """
    Return a loaded ELF header object pointing to the Ehdr of the
    main executable.
    """
    e = entry()
    if e:
        return load(e)
    return None


@pwndbg.aglib.proc.OnlyWhenRunning
@pwndbg.lib.cache.cache_until("start", "objfile")
def entry() -> int:
    """
    Return the address of the entry point for the main executable.
    """
    entry = pwndbg.auxv.get().AT_ENTRY
    if entry:
        return entry

    inf = pwndbg.dbg.selected_inferior()
    entry = inf.main_module_entry()
    if entry:
        return entry

    # Try common names
    for name in ["_start", "start", "__start", "main"]:
        try:
            return pwndbg.aglib.symbol.lookup_symbol_addr(name) or 0
        except pwndbg.dbg_mod.Error:
            pass

    # Can't find it, give up.
    return 0


def load(pointer: int) -> Ehdr | None:
    return get_ehdr(pointer)[1]


ehdr_type_loaded = 0


@pwndbg.lib.cache.cache_until("start", "objfile")
def reset_ehdr_type_loaded() -> None:
    global ehdr_type_loaded
    ehdr_type_loaded = 0


def get_ehdr(pointer: int) -> tuple[int | None, Ehdr | None]:
    """
    Returns an ehdr object for the ELF pointer points into.

    We expect the `pointer` to be an address from the binary.
    """

    base = None

    if pwndbg.aglib.qemu.is_qemu():
        # Only check if the beginning of the page contains the ELF magic,
        # since we cannot get the memory map in qemu-user.
        page_start = pwndbg.lib.memory.page_align(pointer)
        if pwndbg.aglib.memory.read(page_start, 4, partial=True) == b"\x7fELF":
            base = page_start
        else:
            return None, None
    else:
        vmmap = pwndbg.aglib.vmmap.find(pointer)

        # If there is no vmmap for the requested address, we can't do much
        # (e.g. it could have been unmapped for whatever reason)
        if vmmap is None:
            return None, None

        # We first check if the beginning of the page contains the ELF magic
        if pwndbg.aglib.memory.read(vmmap.start, 4, partial=True) == b"\x7fELF":
            base = vmmap.start

        # The page did not have ELF magic; it may be that .text and binary start are split
        # into two pages, so let's get the first page from the pointer's page objfile
        else:
            for v in pwndbg.aglib.vmmap.get():
                if v.objfile == vmmap.objfile:
                    vmmap = v
                    break

            if pwndbg.aglib.memory.read(vmmap.start, 4, partial=True) == b"\x7fELF":
                base = vmmap.start

    if base is None:
        # For non linux ABI, the ELF header may not exist at all
        if pwndbg.dbg.selected_inferior().is_linux():
            print("ERROR: Could not find ELF base!")
        return None, None

    # Determine whether it's 32- or 64-bit
    ei_class = pwndbg.aglib.memory.byte(base + 4)

    # Find out where the section headers start
    Elfhdr: Elf32_Ehdr | Elf64_Ehdr | None = read(Ehdr, base)  # type: ignore[type-var]
    return ei_class, Elfhdr


def get_phdrs(pointer: int):
    """
    Returns a tuple containing (phnum, phentsize, gdb.Value),
    where the gdb.Value object is an ELF Program Header with
    the architecture-appropriate structure type.
    """
    _, Elfhdr = get_ehdr(pointer)

    if Elfhdr is None:
        return (0, 0, None)

    phnum = Elfhdr.e_phnum
    phoff = Elfhdr.e_phoff
    phentsize = Elfhdr.e_phentsize

    x = (phnum, phentsize, read(Phdr, Elfhdr.address + phoff))  # type: ignore[type-var]
    return x


def iter_phdrs(ehdr: Ehdr):
    if not ehdr:
        return

    phnum, phentsize, phdr = get_phdrs(ehdr.address)

    if not phdr:
        return

    first_phdr = phdr.address
    PhdrType = phdr.type

    for i in range(0, phnum):
        p_phdr = int(first_phdr + (i * phentsize))
        p_phdr = read(PhdrType, p_phdr)
        yield p_phdr


def map(pointer: int, objfile: str = "") -> tuple[pwndbg.lib.memory.Page, ...]:
    """
    Given a pointer into an ELF module, return a list of all loaded
    sections in the ELF.

    Returns:
        A sorted list of pwndbg.lib.memory.Page objects

    Example:

        >>> pwndbg.aglib.elf.load(pwndbg.aglib.regs.pc)
        [Page('400000-4ef000 r-xp 0'),
         Page('6ef000-6f0000 r--p ef000'),
         Page('6f0000-6ff000 rw-p f0000')]
        >>> pwndbg.aglib.elf.load(0x7ffff77a2000)
        [Page('7ffff75e7000-7ffff77a2000 r-xp 0x1bb000 0'),
         Page('7ffff77a2000-7ffff79a2000 ---p 0x200000 1bb000'),
         Page('7ffff79a2000-7ffff79a6000 r--p 0x4000 1bb000'),
         Page('7ffff79a6000-7ffff79ad000 rw-p 0x7000 1bf000')]
    """
    ei_class, ehdr = get_ehdr(pointer)
    return map_inner(ei_class, ehdr, objfile)


def map_inner(ei_class: int, ehdr: Ehdr, objfile: str) -> tuple[pwndbg.lib.memory.Page, ...]:
    if not ehdr:
        return ()

    base = int(ehdr.address)
    ptrsize: int = pwndbg.aglib.arch.ptrsize

    # For each Program Header which would load data into our
    # address space, create a representation of each individual
    # page and its permissions.
    #
    # Entries are processed in-order so that later entries
    # which change page permissions (e.g. PT_GNU_RELRO) will
    # override their small subset of address space.
    pages: list[pwndbg.lib.memory.Page] = []
    for phdr in iter_phdrs(ehdr):
        memsz = int(phdr.p_memsz)

        if not memsz:
            continue

        vaddr = int(phdr.p_vaddr)
        offset = int(phdr.p_offset)
        flags = int(phdr.p_flags)

        memsz += pwndbg.lib.memory.page_offset(vaddr)
        memsz = pwndbg.lib.memory.page_size_align(memsz)
        vaddr = pwndbg.lib.memory.page_align(vaddr)
        offset = pwndbg.lib.memory.page_align(offset)

        # For each page described by this program header
        for page_addr in range(vaddr, vaddr + memsz, pwndbg.lib.memory.PAGE_SIZE):
            if page_addr in pages:
                page = pages[pages.index(page_addr)]  # type: ignore[arg-type]

                # Don't ever remove the execute flag.
                # Sometimes we'll load a read-only area into .text
                # and the loader doesn't actually *remove* the executable flag.
                if page.flags & PF_X:
                    flags |= PF_X
                page.flags = flags
            else:
                page = pwndbg.lib.memory.Page(
                    page_addr,
                    pwndbg.lib.memory.PAGE_SIZE,
                    flags,
                    offset + (page_addr - vaddr),
                    ptrsize,
                )
                pages.append(page)

    # Adjust against the base address that we discovered
    # for binaries that are relocatable / type DYN.
    if ET_DYN == int(ehdr.e_type):
        for page in pages:
            page.vaddr += base

    # Merge contiguous sections of memory together
    pages.sort()
    prev = pages[0]
    for page in list(pages[1:]):
        if (prev.flags & PF_W) == (page.flags & PF_W) and prev.vaddr + prev.memsz == page.vaddr:
            prev.memsz += page.memsz
            pages.remove(page)
        else:
            prev = page

    # Fill in any gaps with no-access pages.
    # This is what the linker does, and what all the '---p' pages are.
    gaps: list[pwndbg.lib.memory.Page] = []
    for i in range(len(pages) - 1):
        a, b = pages[i : i + 2]
        a_end = a.vaddr + a.memsz
        b_begin = b.vaddr
        if a_end != b_begin:
            gaps.append(pwndbg.lib.memory.Page(a_end, b_begin - a_end, 0, b.offset, ptrsize))

    pages.extend(gaps)

    for page in pages:
        page.objfile = objfile

    return tuple(sorted(pages))


gcc_compiler_path = pwndbg.config.add_param(
    "gcc-compiler-path",
    "",
    "path to the gcc/g++ toolchain for generating imported symbols",
    param_class=pwndbg.lib.config.PARAM_OPTIONAL_FILENAME,
)


def compile_with_flags(compiler_flags: list[str]) -> Status:
    """
    Compile a C program.

    If the `gcc_compiler_path` argument is set, gcc will be used, otherwise
    zig (the python package) will be used.

    Arguments:
        compiler_flags: The flags to pass to the compiler, including the input and
            output files.

    Returns:
        A status object carrying an error message if compilation failed.
    """
    if gcc_compiler_path != "":
        compiler_cmdline = [str(gcc_compiler_path)]
    else:
        try:
            compiler_cmdline = pwndbg.lib.zig.flags(pwndbg.aglib.arch)
        except ValueError as exception:
            return Status.fail(str(exception))

    gcc_cmd: list[str] = compiler_cmdline + compiler_flags

    try:
        # capture_output=True makes it so the compilation errors are not instantly
        # dumped to the user, but are in the CalledProcessError object.
        # https://docs.python.org/3/library/subprocess.html#subprocess.run:~:text=stdout%20and%20stderr%20if%20they%20were%20captured
        subprocess.run(gcc_cmd, check=True, text=True, capture_output=True)
        return Status()
    except subprocess.CalledProcessError as exception:
        return Status.fail(
            str(exception)
            + f"\nStdout: {exception.stdout}"
            + f"\nStderr: {exception.stderr}"
            + f"\nFailed to compile {compiler_flags[0]}. Please fix any compilation errors there may be."
        )
    except Exception as exception:
        return Status.fail(str(exception) + "\nAn error occurred while compiling.")



"""
QEMU-related utility functions.
"""

from __future__ import annotations

import re

_QEMU_VERSION_RE = re.compile(r"(\d+)\.(\d+)(?:\.(\d+))?")


def parse_qgdbserverversion(response: bytes) -> tuple[int, ...] | None:
    """
    Parse the response from qGDBServerVersion packet and extract version tuple.

    Args:
        response: Raw bytes response from the qGDBServerVersion packet

    Returns:
        Tuple of version integers (e.g., (10, 1, 0)) or None if parsing fails
    """
    if not response or response.startswith(b"E"):
        return None

    text = response.decode(errors="ignore").strip()
    if not text:
        return None

    match = _QEMU_VERSION_RE.search(text)
    if not match:
        return None

    return tuple(int(part) for part in match.groups() if part is not None)



"""
Perform queries specific to the GNU C Library.
"""

from __future__ import annotations

import re

import pwndbg.aglib.elf
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.lib.cache
import pwndbg.lib.config
from pwndbg.color import message
from pwndbg.lib.config import Scope

from . import util
from .dispatch import LibcType
from .dispatch import LibcURLs

safe_lnk = pwndbg.config.add_param(
    "safe-linking",
    None,
    "whether glibc uses safe-linking",
    param_class=pwndbg.lib.config.PARAM_AUTO_BOOLEAN,
)

glibc_version = pwndbg.config.add_param(
    "glibc", "", "glibc version for heap heuristics resolution (e.g. 2.31)", scope=Scope.heap
)


@pwndbg.config.trigger(glibc_version)
def set_glibc_version() -> None:
    ret = re.search(r"^(\d+)\.(\d+)$", glibc_version.value)
    if ret:
        return

    print(
        message.warn(
            f"Invalid GLIBC version: `{glibc_version.value}`, you should provide something like: 2.31 or 2.34"
        )
    )
    glibc_version.revert_default()


@pwndbg.lib.cache.cache_until("start", "objfile")
def _get_version(libc_filepath: str) -> tuple[int, ...]:
    addr = pwndbg.aglib.symbol.lookup_symbol_addr("__libc_version", objfile_endswith=libc_filepath)
    if addr is not None:
        ver = pwndbg.aglib.memory.string(addr)
        return util.version_parse(ver)

    result = pwndbg.aglib.elf.section_by_name(libc_filepath, ".rodata", try_local_path=True)
    if result is None:
        raise Exception(f"Could not retrieve .rodata section of glibc {libc_filepath}")

    _, _, data = result
    banner_start = data.find(b"GNU C Library")
    if banner_start == -1:
        raise Exception(
            f"Could not find 'GNU C Library' in .rodata section of glibc {libc_filepath}"
        )

    banner = data[banner_start : data.find(b"\x00", banner_start)]
    ret = re.search(rb"release version (\d+)\.(\d+)", banner)
    if ret is None:
        raise Exception(
            f"Could not find 'release version' in .rodata section of glibc {libc_filepath}"
        )

    return tuple(int(_) for _ in ret.groups())


# ===== Libc Interface Implementation =====


def type() -> LibcType:
    return LibcType.GLIBC


def version(libc_filepath: str) -> tuple[int, ...]:
    if glibc_version:
        version_tuple = tuple(int(i) for i in glibc_version.value.split("."))
        return version_tuple

    return _get_version(libc_filepath)


# NOTE: Operating under the assumption that debuginfod and add-symbol-file
# trigger the objfile event, making it safe to cache these functions like this.


@pwndbg.lib.cache.cache_until("start", "objfile")
def has_internal_symbols(libc_filepath: str) -> bool:
    # __libc_version exists in all versions of glibc and is an internal symbol.
    # https://elixir.bootlin.com/glibc/glibc-1.90/source/version.c#L23
    return (
        pwndbg.aglib.symbol.lookup_symbol("__libc_version", objfile_endswith=libc_filepath)
        is not None
    )


@pwndbg.lib.cache.cache_until("start", "objfile")
def has_debug_info() -> bool:
    return pwndbg.aglib.typeinfo.load("struct malloc_chunk") is not None


def urls(ver: tuple[int, ...] | None) -> LibcURLs:
    assert ver is not None
    ver_str = ".".join(map(str, ver))
    return LibcURLs(
        versioned_readable_source=f"https://elixir.bootlin.com/glibc/glibc-{ver_str}/source",
        versioned_compressed_source=f"https://ftp.gnu.org/gnu/libc/glibc-{ver_str}.tar.gz",
        homepage="https://sourceware.org/glibc/",
        git="https://sourceware.org/git/glibc.git",
    )


def verify_libc_candidate(mapping_name: str) -> bool:
    if has_internal_symbols(mapping_name):
        # __GI_exit exists since at least version 2.3 (until at least 2.42)
        # https://elixir.bootlin.com/glibc/glibc-2.3/source/include/libc-symbols.h#L670
        # https://elixir.bootlin.com/glibc/glibc-2.3/source/include/libc-symbols.h#L642
        # https://elixir.bootlin.com/glibc/glibc-2.3/source/stdlib/exit.c#L84
        # and I don't see it in other libc's.
        return (
            pwndbg.aglib.symbol.lookup_symbol("__GI_exit", objfile_endswith=mapping_name)
            is not None
        )
    # We don't have internal symbols so we will use a more expensive (?) check:
    rodata: tuple[int, int, bytes] | None = pwndbg.aglib.elf.section_by_name(
        mapping_name, ".rodata", try_local_path=True
    )
    if rodata is None:
        return False
    _, _, data = rodata

    return b"GNU C Library" in data


def verify_ld_candidate(mapping_name: str) -> bool:
    return False


def libc_same_as_ld() -> bool:
    return False


# ===== End of Libc Interface Implementation =====


def check_safe_linking(ver: tuple[int, ...]) -> bool:
    """
    Arguments:
        ver: The version tuple. Pass pwndbg.libc.version() here.

    Safe-linking is a glibc 2.32 mitigation; see:
    - https://lanph3re.blogspot.com/2020/08/blog-post.html
    - https://research.checkpoint.com/2020/safe-linking-eliminating-a-20-year-old-malloc-exploit-primitive/
    """
    # FIXME: What if we are not being used?
    return (ver >= (2, 32) or safe_lnk.value) and safe_lnk.value is not False
